import numpy as np
import scipy.special as sp

from UncertainSCI.opoly1d import eval_driver, \
        leading_coefficient_driver, gauss_quadrature_driver

# from UncertainSCI.utils.compute_subintervals import compute_subintervals
from UncertainSCI.utils.quad import gq_modification_composite, \
        gq_modification_unbounded_composite, compute_subintervals

"""
Predictor-corrector method
"""


def predict_correct_bounded(a, b, weight, N, singularity_list, Nquad=10):
    """ Three-term recurrence coefficients from quadrature

    Computes the first N three-term recurrence coefficient pairs associated to
    weight on the bounded interval [a, b].

    Performs global integration on [a, b] using
    utils.quad.gq_modification_composite.
    """

    assert a < b

    # First divide [a, b] into subintervals based on singularity locations.
    subintervals = compute_subintervals(a, b, singularity_list)

    ab = np.zeros([N, 2])
    ab[0, 1] = np.sqrt(gq_modification_composite(weight, a, b, Nquad,
                       subintervals=subintervals))

    peval = lambda x, n: eval_driver(x, np.array([n]), 0, ab)

    for n in range(0, N-1):
        # Guess next coefficients
        ab[n+1, 0], ab[n+1, 1] = ab[n, 0], ab[n, 1]

        integrand = lambda x: weight(x) * peval(x, n).flatten() *\
                      peval(x, n+1).flatten()
        ab[n+1, 0] += ab[n, 1] * gq_modification_composite(integrand, a, b,
                                                           n+1+Nquad,
                                                           subintervals)

        integrand = lambda x: weight(x) * peval(x, n+1).flatten()**2
        ab[n+1, 1] *= np.sqrt(gq_modification_composite(integrand, a, b,
                                                        n+1+Nquad,
                                                        subintervals))

    return ab


def predict_correct_unbounded(a, b, weight, N, singularity_list, Nquad=10):
    assert a < b
    ab = np.zeros([N, 2])
    ab[0, 1] = np.sqrt(gq_modification_unbounded_composite(weight, a, b, Nquad,
                                                           singularity_list))

    peval = lambda x, n: eval_driver(x, np.array([n]), 0, ab)

    for n in range(0, N-1):
        # Guess next coefficients
        ab[n+1, 0], ab[n+1, 1] = ab[n, 0], ab[n, 1]

        integrand = lambda x: weight(x) * peval(x, n).flatten() *\
                              peval(x, n+1).flatten()
        ab[n+1, 0] += ab[n, 1] * gq_modification_unbounded_composite(integrand,
                                    a, b, n+1+Nquad, singularity_list)

        integrand = lambda x: weight(x) * peval(x, n+1).flatten()**2
        ab[n+1, 1] *= np.sqrt(gq_modification_unbounded_composite(integrand, a,
                                b, n+1+Nquad, singularity_list))

    return ab


def predict_correct_discrete(xg, wg, N):

    assert all(i >= 0 for i in wg)
    assert N <= len(xg)

    ab = np.zeros([N, 2])
    ab[0, 1] = np.sqrt(np.sum(wg))

    peval = lambda x, n: eval_driver(x, np.array([n]), 0, ab)

    for n in range(0, N-1):
        # Guess next coefficients
        ab[n+1, 0], ab[n+1, 1] = ab[n, 0], ab[n, 1]

        integrand = lambda x: peval(x, n).flatten() * peval(x, n+1).flatten()
        ab[n+1, 0] += ab[n, 1] * np.sum(integrand(xg) * wg)

        integrand = lambda x: peval(x, n+1).flatten()**2
        ab[n+1, 1] *= np.sqrt(np.sum(integrand(xg) * wg))

    return ab


def predict_correct_bounded_composite(a, b, weight, N, singularity_list,
                                      Nquad=10):
    """ Three-term recurrence coefficients from composite quadrature

    Computes the first N three-term recurrence coefficient pairs associated to
    weight on the bounded interval [a, b]. 

    Performs composite quadrature on [a, b] using
    utils.quad.gq_modification_composite.
    """

    assert a < b

    # First divide [a, b] into subintervals based on singularity locations.
    global_subintervals = compute_subintervals(a, b, singularity_list)

    ab = np.zeros([N, 2])
    ab[0, 1] = np.sqrt(gq_modification_composite(weight, a, b, Nquad, 
                        subintervals=global_subintervals))

    integrand = weight

    for n in range(0, N-1):
        # Guess next coefficients
        ab[n+1, 0], ab[n+1, 1] = ab[n, 0], ab[n, 1]

        # Set up linear modification roots and subintervals
        breaks = singularity_list.copy()
        pn_zeros = gauss_quadrature_driver(ab, n)[0]
        pn1_zeros = gauss_quadrature_driver(ab, n+1)[0]

        roots = np.hstack([pn_zeros, pn1_zeros])
        breaks += [[z, 0, 0] for z in roots]
        subintervals = compute_subintervals(a, b, breaks)

        # Leading coefficient
        qlc = np.prod(leading_coefficient_driver(n+2, ab)[-2:])

        ab[n+1, 0] += ab[n, 1] * gq_modification_composite(integrand, a, b,
                                                           n+1+Nquad,
                                                           subintervals,
                                                           roots=roots,
                                                           leading_coefficient=qlc)

        # Here subintervals are the global ones
        pn1_zeros = gauss_quadrature_driver(ab, n+1)[0]
        qlc = (leading_coefficient_driver(n+2, ab)[-1])**2

        ab[n+1, 1] *= np.sqrt(gq_modification_composite(integrand, a, b,
                                n+1+Nquad, global_subintervals,
                                quadroots=pn1_zeros, leading_coefficient=qlc))

    return ab

def predict_correct_unbounded_composite(a, b, weight, N, singularity_list,
                                        Nquad=10):
    assert a < b

    ab = np.zeros([N, 2])
    ab[0, 1] = np.sqrt(gq_modification_unbounded_composite(weight, a, b, Nquad,
                                                           singularity_list))

    integrand = weight

    for n in range(0, N-1):
        # Guess next coefficients
        ab[n+1, 0], ab[n+1, 1] = ab[n, 0], ab[n, 1]

        # Set up linear modification roots and subintervals
        breaks = singularity_list.copy()
        pn_zeros = gauss_quadrature_driver(ab, n)[0]
        pn1_zeros = gauss_quadrature_driver(ab, n+1)[0]

        roots = np.hstack([pn_zeros, pn1_zeros])
        breaks += [[z, 0, 0] for z in roots]

        # Leading coefficient
        qlc = np.prod(leading_coefficient_driver(n+2, ab)[-2:])

        ab[n+1, 0] += ab[n, 1] * gq_modification_unbounded_composite(integrand,
                                    a, b, n+1+Nquad, breaks, roots=roots,
                                    leading_coefficient=qlc)

        # Here subintervals are the global ones
        pn1_zeros = gauss_quadrature_driver(ab, n+1)[0]
        qlc = (leading_coefficient_driver(n+2, ab)[-1])**2

        ab[n+1, 1] *= np.sqrt(gq_modification_unbounded_composite(integrand, a,
                                b, n+1+Nquad, singularity_list,
                                quadroots=pn1_zeros, leading_coefficient=qlc))

    return ab


"""
Stieltjes procedure
"""
def stieltjes_bounded(a, b, weight, N, singularity_list, Nquad=10):


    assert a < b

    subintervals = compute_subintervals(a, b, singularity_list)

    ab = np.zeros([N, 2])
    ab[0, 1] = np.sqrt(gq_modification_composite(weight, a, b, Nquad,
                                                 subintervals=subintervals))

    peval = lambda x, n: eval_driver(x, np.array([n]), 0, ab)

    for n in range(1, N):
        integrand = lambda x: weight(x) * x * peval(x, n-1).flatten()**2
        ab[n, 0] = gq_modification_composite(integrand, a, b, n+1+Nquad,
                                             subintervals)
        if n == 1:
            integrand = lambda x: weight(x) * ((x - ab[n, 0]) *
                                  peval(x, n-1).flatten())**2
        else:
            integrand = lambda x: weight(x) * ((x - ab[n, 0]) *
                                  peval(x, n-1).flatten() - ab[n-1, 1] *
                                  peval(x, n-2).flatten())**2
        ab[n, 1] = np.sqrt(gq_modification_composite(integrand, a, b,
                                                      n+1+Nquad, subintervals))

    return ab

def stieltjes_unbounded(a, b, weight, N, singularity_list, Nquad=10):

    assert a < b

    ab = np.zeros([N, 2])
    ab[0, 1] = np.sqrt(gq_modification_unbounded_composite(weight, a, b, Nquad,
                                                           singularity_list))

    peval = lambda x, n: eval_driver(x, np.array([n]), 0, ab)

    for n in range(1, N):
        integrand = lambda x: weight(x) * x * peval(x, n-1).flatten()**2
        ab[n, 0] = gq_modification_unbounded_composite(integrand, a, b,
                                                       n+1+Nquad,
                                                       singularity_list)
        if n == 1:
            integrand = lambda x: weight(x) * ((x - ab[n, 0]) *\
                                  peval(x, n-1).flatten())**2
        else:
            integrand = lambda x: weight(x) * ((x - ab[n, 0]) *\
                                  peval(x, n-1).flatten() - ab[n-1, 1] *\
                                  peval(x, n-2).flatten())**2
        ab[n, 1] = np.sqrt(gq_modification_unbounded_composite(integrand, a,
                                b, n+1+Nquad, singularity_list))

    return ab

def stieltjes_discrete(xg, wg, N):

    assert all(i >=0 for i in wg)
    assert N <= len(xg)

    ab = np.zeros([N, 2])
    ab[0, 1] = np.sqrt(np.sum(wg))

    peval = lambda x, n: eval_driver(x, np.array([n]), 0, ab)

    for n in range(1, N):
        integrand = lambda x: x * peval(x, n-1).flatten()**2
        ab[n, 0] = np.sum(integrand(xg) * wg)
        if n == 1:
            integrand = lambda x: ((x - ab[n, 0]) *\
                                    peval(x, n-1).flatten())**2
        else:
            integrand = lambda x: ((x - ab[n, 0]) * peval(x, n-1).flatten() -\
                                    ab[n-1, 1] * peval(x, n-2).flatten())**2
        ab[n, 1] = np.sqrt(np.sum(integrand(xg) * wg))

    return ab

def stieltjes_bounded_composite(a, b, weight, N, singularity_list, Nquad=10):

    assert a < b

    # First divide [a, b] into subintervals based on singularity locations.
    global_subintervals = compute_subintervals(a, b, singularity_list)

    ab = np.zeros([N, 2])
    ab[0, 1] = np.sqrt(gq_modification_composite(weight, a, b, Nquad, 
                                             subintervals=global_subintervals))

    integrand = weight

    for n in range(1, N):

        breaks = singularity_list.copy()
        pnminus1_zeros = gauss_quadrature_driver(ab, n-1)[0]
        roots = np.hstack([0, pnminus1_zeros])
        breaks += [[z, 0, 0] for z in roots]
        subintervals = compute_subintervals(a, b, breaks)

        qlc = np.prod(leading_coefficient_driver(n, ab)[-1])**2

        ab[n, 0] = gq_modification_composite(integrand, a, b, n+1+Nquad,
                                             subintervals, roots=np.zeros(1,),
                                             quadroots=pnminus1_zeros,
                                             leading_coefficient=qlc)

        if n == 1:
            pnminus1_zeros = np.hstack([ab[n, 0], pnminus1_zeros])
            s = gq_modification_composite(integrand, a, b, n+1+Nquad,
                                          global_subintervals,
                                          quadroots=pnminus1_zeros,
                                          leading_coefficient=qlc)
        else:
            pnminus1_zeros = np.hstack([ab[n, 0], pnminus1_zeros])
            s_1 = gq_modification_composite(integrand, a, b, n+1+Nquad,
                                            global_subintervals,
                                            quadroots=pnminus1_zeros,
                                            leading_coefficient=qlc)

            pnminus2_zeros = gauss_quadrature_driver(ab, n-2)[0]
            qlc = np.prod(leading_coefficient_driver(n-1, ab)[-1])**2
            s_2 = gq_modification_composite(integrand, a, b, n+1+Nquad,
                                            global_subintervals,
                                            quadroots=pnminus2_zeros,
                                            leading_coefficient=qlc)

            # below pnminus1_zeros already includes ab[n, 0]
            roots = np.hstack([pnminus1_zeros, pnminus2_zeros]) 
            breaks_new = singularity_list.copy()
            breaks_new += [[z, 0, 0] for z in roots]
            subintervals = compute_subintervals(a, b, breaks_new)
            qlc = np.prod(leading_coefficient_driver(n, ab)[-2:])
            s_3 = gq_modification_composite(integrand, a, b, n+1+Nquad,
                                            subintervals, roots=roots,
                                            leading_coefficient=qlc)

            s = s_1 + ab[n-1, 1]**2 * s_2 - 2 * ab[n-1, 1] * s_3

        ab[n, 1] = np.sqrt(s)

    return ab

def stieltjes_unbounded_composite(a, b, weight, N, singularity_list, Nquad=10):

    assert a < b

    ab = np.zeros([N, 2])
    ab[0, 1] = np.sqrt(gq_modification_unbounded_composite(weight, a, b, Nquad,
                                                           singularity_list))

    integrand = weight

    for n in range(1, N):

        breaks = singularity_list.copy()
        pnminus1_zeros = gauss_quadrature_driver(ab, n-1)[0]
        roots = np.hstack([0, pnminus1_zeros])
        # use array_unique because when n = 2, pnminus1_zeros = ab[1, 0] is
        # very close to 0.

        # Have to do this array_unique, or will cause Overlapping singularities
        # problem.
        breaks += [[z, 0, 0] for z in roots]

        qlc = np.prod(leading_coefficient_driver(n, ab)[-1])**2

        ab[n, 0] = gq_modification_unbounded_composite(integrand, a, b,
                                                       n+1+Nquad, breaks,
                                                       roots=np.zeros(1,),
                                                       quadroots=pnminus1_zeros,
                                                       leading_coefficient=qlc)

        if n == 1:
            pnminus1_zeros = np.hstack([ab[n, 0], pnminus1_zeros])
            s = gq_modification_unbounded_composite(integrand, a, b, n+1+Nquad,
                                                    singularity_list,
                                                    quadroots=pnminus1_zeros,
                                                    leading_coefficient=qlc)
        else:
            pnminus1_zeros = np.hstack([ab[n, 0], pnminus1_zeros])
            s_1 = gq_modification_unbounded_composite(integrand, a, b, n+1+Nquad,
                                                      singularity_list,
                                                      quadroots=pnminus1_zeros,
                                                      leading_coefficient=qlc)

            pnminus2_zeros = gauss_quadrature_driver(ab, n-2)[0]
            qlc = np.prod(leading_coefficient_driver(n-1, ab)[-1])**2
            s_2 = gq_modification_unbounded_composite(integrand, a, b,
                                                      n+1+Nquad,
                                                      singularity_list,
                                                      quadroots=pnminus2_zeros,
                                                      leading_coefficient=qlc)

            # note this roots may cause issues since it contains two very close
            # numbers
            # below pnminus1_zeros already includes ab[n, 0]
            roots = np.hstack([pnminus1_zeros, pnminus2_zeros]) 
            breaks_new = singularity_list.copy()
            breaks_new += [[z, 0, 0] for z in roots]
            qlc = np.prod(leading_coefficient_driver(n, ab)[-2:])
            s_3 = gq_modification_unbounded_composite(integrand, a, b,
                                                      n+1+Nquad, breaks_new,
                                                      roots=roots,
                                                      leading_coefficient=qlc)

            s = s_1 + ab[n-1, 1]**2 * s_2 - 2 * ab[n-1, 1] * s_3

        ab[n, 1] = np.sqrt(s)

    return ab



"""
Arbitrary polynomial chaos expansion method
"""
def expansion_coeff(m, n):
    """
    Params
    ______
    m: numpy array
    moments wst given measure
    M = m_0     ... m_n
        .
        .
        m_{n-1} ... m_{2n-1}
        0       ... 1
    requires [m_0,  ...m_{2n-1}], i.e. len(m) >= 2n

    n: int
    the highest order of polynomials expansion
    p_n(x) = \sum_{i=0}^n c_i^(n) x^i

    Returns
    ------
    c: numpy array, shape (n+1,)
    vector of expansion coeffcients for monic polynomials,
    c = [c_0^(n), ..., c_n^(n)]
    """
    assert len(m) >= 2*n
    M = np.zeros([n+1, n+1])
    for i in range(n):
        M[i, :] = m[i:i+n+1]
    M[n, n] = 1.
    b = np.zeros(n+1,)
    b[n] = 1
    c = np.linalg.solve(M, b)
    return c

def normal_const(m, n, c):
    """
    Params
    ------
    m: numpy array
    moments wst given measure
    M = m_0     ... m_n
        .
        .
        m_{n-1} ... m_{2n-1}
        m_n     ... m_{2n} 
    requires [m_0, ...m_{2n}], i.e. len(m) >= 2n+1

    n: int
    the highest order of polynomials expansion
    p_n(x) = \sum_{i=0}^n c_i^(n) x^i

    c: numpy array, shape (n+1)
    vector of expansion coeffcients for monic polynomials
    c = [c_0^(n), ..., c_n^(n)]

    Returns
    ------
    normal_c: float
    normalized constant for expansion coefficient vector c
    """
    assert len(m) >= 2*n+1
    M = np.zeros([n+1, n+1])
    for i in range(n+1):
        M[i, :] = m[i:i+n+1]
    normal_c = np.sqrt(c.dot(M).dot(c))
    return normal_c

def aPC(m, N):
    C = np.zeros([N, N])
    NC = np.zeros(N,)
    for i in range(N):
        c = expansion_coeff(m, i)
        NC[i] = normal_const(m, i, c)
        C[0:i+1, i] = expansion_coeff(m, i) / NC[i]

    ab = np.zeros([N, 2])
    ab[0, 1] = NC[0]
    ab[1, 1] = C[0, 0] / C[1, 1]
    ab[1, 0] = -C[0, 1] / C[1, 1]
    for i in range(2, N):
        ab[i, 1] = C[i-1, i-1] / C[i, i]
        ab[i, 0] = (C[i-2, i-1] - ab[i, 1] * C[i-1, i]) / C[i-1, i-1]

    return ab



"""
Hankel determinants method (Classic moment method)
"""
def deter(m, n):
    """
    compute the Hankel determinant of order n, H_n
    in the moments m_0, ..., m_{2n-2},
    i.e. the first 2n-1 moments are used.
    """
    assert len(m) >= 2*n-1, 'need more moments'
    if n == 0:
        return 1
    elif n == 1:
        return m[0]
    else:
        A = np.zeros((n, n))
        for i in range(n):
            A[i, :] = m[i:i+n]
        assert A.shape == (n, n)
    return np.linalg.det(A)

def deter_mod(m, n):
    """
    compute the Hankel determinant of order n+1, H_{n+1}
    with the penultimate column and the last row removed,
    in the moments m_0, ..., m_{2n},
    i.e. the first 2n+1 moments are used.
    """
    assert len(m) >= 2*n+1, 'need more moments'
    if n == 0:
        return 0
    elif n == 1:
        return m[1]
    else:
        A = np.zeros((n+1, n+1))
        for i in range(n+1):
            A[i, :] = m[i:i+n+1]
        B = np.delete(A, -1, axis = 0)
        B = np.delete(B, -2, axis = 1)
        assert B.shape == (n, n)
    return np.linalg.det(B)

def hankel_deter(N, m):
    assert len(m) >= 2*N-1, 'need more moments'
    ab = np.zeros([N, 2])
    ab[0, 1] = m[0]
    for i in range(1, N):
        ab[i, 0] = deter_mod(m, i) / deter(m, i) \
                - deter_mod(m, i-1) / deter(m, i-1)
        ab[i, 1] = deter(m, i+1) * deter(m, i-1) / deter(m, i)**2
    ab[:, 1] = np.sqrt(ab[:, 1])
    return ab



"""
Modified Chebyshev Algorithm
"""
def mod_cheb(N, mod_m, lbd):
    """ compute the first N recurrence coefficients wrt d\mu

    Params:
    d\\mu: given measure,
    mod_m: the first 2N-1 modified moments mod_m_1, ..., mod_m_{2N-2}
             can be computed by quad.gq_modification_composite
    lbd: known measure to be chosen s.t. d\\lbd ~ d\\mu in some sense
         compute the first 2N recurrence coefficients {a_k, b_k}_{k=0}^{2N-1}
    """
    assert len(mod_m) == 2*N-1, 'Need more modified moments'

    albe = lbd.recurrence(N=2*N-1)
    al, be = albe[:, 0], albe[:, 1]

    sigma = np.zeros([N, 2*N-1])
    sigma[0, :] = mod_m

    a_1 = al[1] + be[1] * mod_m[1] / mod_m[0]
    for k in range(1, 2*N-2):
        sigma[1, k] = be[k] * sigma[0, k-1] + \
                (al[k+1] - a_1) * sigma[0, k] + be[k+1] * sigma[0, k+1]

    for n in range(2, N):
        for k in range(n, 2*N-n-1):
            sigma[n, k] = be[k] * sigma[n-1, k-1] + \
                (al[k+1] - (al[n] + be[n] * sigma[n-1, n] / sigma[n-1, n-1] - \
                be[n-1] * sigma[n-2, n-1] / sigma[n-2, n-2])) * sigma[n-1, k] + \
                be[k+1] * sigma[n-1, k+1] - \
                (be[n-1] * sigma[n-1, n-1] / sigma[n-2, n-2]) * sigma[n-2, k]

    ab = np.zeros([N, 2])
    ab[0, 1] = mod_m[0] * be[0]
    ab[1, 0] = a_1
    for j in range(2, N):
        ab[j, 0] = al[j] + be[j] * sigma[j-1, j] / sigma[j-1, j-1] - \
                be[j-1] * sigma[j-2, j-1] / sigma[j-2, j-2]

        ab[j-1, 1] = be[j-1] * sigma[j-1, j-1] / sigma[j-2, j-2]

    ab[N-1, 1] = be[N-1] * sigma[N-1, N-1] / sigma[N-2, N-2]
    ab[:, 1] = np.sqrt(ab[:, 1])
    return ab



"""
Discrete Painlevé I equation method
"""
def delta(n):
        return (1 - (-1)**n) / 2

def dPI4(N, rho=0.):
    ab = np.zeros((N, 2))
    ab[0, 1] = (1/2) * sp.gamma((1+rho)/4)
    ab[1, 1] = (1/2) * sp.gamma((3+rho)/4) / ab[0, 1]

    # x_n = 2 * b_n^2, initial values: x_0 = 0 (b_0 = 0) and x_1
    ab[0, 1] = 0.
    ab[1, 1] = 2 * ab[1, 1]
    for i in range(2, N):
        ab[i, 1] = (i-1 + rho * delta(i-1)) / ab[i-1, 1] - ab[i-1, 1] - ab[i-2, 1]
    ab[:, 1] = np.sqrt(ab[:, 1]/2)
    ab[0, 1] = np.sqrt((1/2) * sp.gamma((1+rho)/4))

    return ab

def dPI6(N, rho = 0.):
    ab = np.zeros((N, 2))
    ab[0, 1] = (1/3) * sp.gamma((1+rho)/6)
    ab[1, 1] = (1/3) * sp.gamma((3+rho)/6) / ab[0, 1]
    ab[2, 1] = 1 / (ab[1, 1]*ab[0, 1]) * (1/3) * sp.gamma((5+rho)/6) \
            - 2 / ab[0, 1] * (1/3) * sp.gamma((3+rho)/6) \
            + ab[1, 1] / ab[0, 1] * (1/3) * sp.gamma((1+rho)/6)
    ab[3, 1] = 1 / (ab[2, 1]*ab[1, 1]*ab[0, 1]) * (1/3) * sp.gamma((7+rho)/6) \
            - 2*(ab[1, 1] + ab[2, 1]) / (ab[2, 1]*ab[1, 1]*ab[0, 1]) * (1/3) * sp.gamma((5+rho)/6) \
            + (ab[1, 1] + ab[2, 1])**2 / (ab[2, 1]*ab[1, 1]*ab[0, 1]) * (1/3) * sp.gamma((3+rho)/6)

    # initial values: x_0 = 0 (b_0 = 0), x_1, x_2 and x_3q
    ab[0, 1] = 0.
    for i in range(4, N):
        ab[i, 1] = ((i-2 + rho * delta(i-2)) / (6 * ab[i-2, 1]) \
                - (ab[i-4, 1]*ab[i-3, 1] + ab[i-3, 1]**2 + 2*ab[i-3, 1]*ab[i-2, 1] + ab[i-3, 1]*ab[i-1, 1] \
                + ab[i-2, 1]**2 + 2*ab[i-2, 1]*ab[i-1, 1] + ab[i-1, 1]**2)) / ab[i-1, 1]

    ab[:, 1] = np.sqrt(ab[:, 1])
    ab[0, 1] = np.sqrt((1/3) * sp.gamma((1+rho)/6))
    return ab



"""
Stabilized Lanczos algorithm
"""
def lanczos_stable(x, w, N):

    assert len(x) == len(w)

    n = len(w) + 1
    w = np.sqrt(w)

    # Initialize variables
    qs = np.zeros([n, n])
    v = np.zeros(n)
    v[0] = 1.
    qs[:, 0] = v

    a = np.zeros(N)
    b = np.zeros(N)

    for s in range(N+1):
        z = np.hstack([v[0] + np.sum(w*v[1:n]), w*v[0] + x * v[1:n]])

        if s > 0:
            a[s-1] = v.dot(z)

        # double orthogonalization
        z = z - qs[:, 0:s+1].dot((qs[:, 0:s+1].T.dot(z)))
        z = z - qs[:, 0:s+1].dot((qs[:, 0:s+1].T.dot(z)))

        if s < N:
            znorm = np.linalg.norm(z)
            b[s] = znorm**2
            v = z / znorm
            qs[:, s+1] = v

    ab = np.zeros([N+1, 2])
    ab[1:, 0] = a
    ab[0:-1, 1] = np.sqrt(b)

    return ab[:-1, :]

# def lanczos_stable(x, w):
#
    # assert len(x) == len(w)
#
    # n = len(w) + 1
    # w = np.sqrt(w)
#
    # qs = np.zeros([n, n])
    # v = np.zeros(n)
    # v[0] = 1.
    # qs[:, 0] = v
#
    # a = np.zeros(n-1)
    # b = np.zeros(n-1)
#
    # for s in range(n):
        # z = np.hstack([ v[0] + np.sum(w*v[1:n]), w*v[0] + x * v[1:n] ])
#
        # if s > 0:
            # a[s-1] = v.dot(z)
#
        # z = z - qs[:, 0:s+1].dot((qs[:, 0:s+1].T.dot(z)))
        # z = z - qs[:, 0:s+1].dot((qs[:, 0:s+1].T.dot(z)))
#
        # if s < n-1:
            # znorm = np.linalg.norm(z)
            # b[s] = znorm**2
            # v = z / znorm
            # qs[:, s+1] = v
#
    # ab = np.zeros([n, 2])
    # ab[1:, 0] = a
    # ab[0:-1, 1] = np.sqrt(b)
#
    # return ab[:-1, :]


# def lanczos(x, w, N):
    # """
    # Given length-M vector x and w, computes the first N coefficients.
    # consider conditions under which the Lanczos algorithm serves as
    # a discrete approximation to the Stieltjes procedure, i.e. stieltjes_discrete
#
    # Requires M >> N
    # """
    # ab = np.zeros([N+1, 2])
    # assert len(x) == len(w), 'x and w should have the same size'
    # assert len(x) >= N, 'size of x should much larger than N'
    # M = len(x)
    # v = np.zeros([M, N])
    # tilde_v = np.zeros([M, N+1])
    # tilde_v[:, 0] = np.sqrt(w)
#
    # for i in range(N):
        # ab[i, 1] = np.linalg.norm(tilde_v[:, i], None)
        # v[:, i] = tilde_v[:, i] / ab[i, 1]
        # ab[i+1, 0] = np.sum(x * v[:, i]**2)
        # if i == 0:
            # tilde_v[:, i+1] = (x - ab[i+1, 0]) * v[:, i]
        # else:
            # tilde_v[:, i+1] = (x - ab[i+1, 0]) * v[:, i] - ab[i, 1] * v[:, i-1]
#
    # return ab[:N]




import numpy as np
from scipy.linalg import qr, qr_delete

from UncertainSCI.utils.version import version_lessthan


def mgs_pivot_restart(A, p=None, pstart=None):
    """
    Computes pivots from a QR decomposition with starting pivots specified. If
    A is an M x N matrix, computes pivots associated to a permutation matrix P
    in a partial QR decomposition of A.T with column pivoting:

      A P = T R,

    where the first p columns of T are orthonormal, and R is an upper
    triangular matrix where the first p rows contain residual entries as in a
    standard QR decomposition. The last N-p rows of R are a slice of the
    identity matrix.

    An ordered list of pivots is returned that are associated to the
    permutation matrix P.

    Args:
        A (numpy array): M x N array
        p (int): The number of pivots to compute. Defaults to None, in which
          case p is set to max(min(M,N), len(pstart)).
        pstart (list/array of ints): Ordered list of user-chosen pivots.
    Returns:
        numpy.ndarray: A vector of ints containing p pivots.
    """

    M, N = A.shape

    if pstart is None:
        pstart = np.zeros(0, dtype=int)
    else:
        assert all(0 <= pval < N for pval in pstart)
        pstart = np.array(pstart, dtype=int)

    p = max(min(M, N), len(pstart))

    # Since we must take pivots in pstart, permute so that these indices are at
    # the top.
    Npst = len(pstart)
    if Npst == 0:
        # More work than necessary, but is probably more efficient
        _, _, inds = qr(A, pivoting=True)
        return inds[:p]

    cpstart = np.setdiff1d(range(N), pstart)
    A = np.hstack((A[:, pstart], A[:, cpstart]))
    inds = np.hstack((pstart, cpstart))

    # Now perform MGS, partly using scipy/lapack routines
    Q, R = qr(A[:, :Npst], mode='economic')

    # MGS Orthogonalization:
    for qq in range(Npst):
        A[:, Npst:] -= np.outer(Q[:, qq], Q[:, qq].T @ A[:, Npst:])

    # Now we just MGS our way to the end.
    for q in range(Npst, p):
        pnext = q + np.argmax(np.sum(A[:, q:]**2, axis=0))

        # Pivot
        inds[[q, pnext]] = inds[[pnext, q]]
        A[:, [q, pnext]] = A[:, [pnext, q]]

        # Orthogonalize
        qvec = A[:, q]
        if np.linalg.norm(qvec) < 1e-13:
            assert False  # Matrix is low-rank so stop pivoting
        qvec /= np.linalg.norm(qvec)
        temp = (qvec.T @ A[:, q+1:])
        A[:, q+1:] -= np.outer(qvec, temp)

    return inds[:p]


def greedy_d_optimal(A, p, pstart=None):
    r"""
    Chooses p rows in A via a greedy D-optimal design. Performs the iterative
    optimization,

        if |R| < A.shape[1]:
            max_r det( A[S, :] * A[S, :].T ),   S = r \cup R,
        else:
            max_r det( A[S, :]^T * A[S, :] ),   S = r \cup R,

    where R = \emptyset is the starting point, and at each step R \gets R \cup
    r^\ast, where r^\ast is the maximizing row.

    If an iterable pstart is given, forces the indices pstart to lie in the set
    R. Returns an error if len(pstart) < N, with N the number of columns of A.

    Returns an ordered pivot vector P indicating the ordered selection of rows
    of A.
    """

    assert len(A.shape) == 2
    if p > A.shape[0]:
        p = A.shape[0]

    M, N = A.shape

    if pstart is None:
        R, P = qr(A.T, pivoting=True, mode='r')
        numpivots = N
    else:
        assert all(0 <= pval <= M-1 for pval in pstart)

        # User asked for fewer pivots than the starting ones
        if len(pstart) >= p:
            return pstart[:p]

        P = np.hstack([np.array(pstart), np.setdiff1d(range(M), pstart)])

        if len(pstart) < N:
            P[:N] = mgs_pivot_restart(A.T, p=N, pstart=pstart)

        numpivots = len(pstart)
        # Otherwise: we have at least N pivots, but fewer than p.

    if p > numpivots:

        W = A[P[:numpivots], :]
        G = np.dot(W.T, W)
        Ginvwm = np.linalg.solve(G, A[P[numpivots:], :].T)

        for m in range(numpivots, p):
            # The remaining choices:
            detnorms = np.sum(A[P[m:], :].T * Ginvwm[:, (m-numpivots):], axis=0)

            # Det maximization
            Pind = np.argmax(detnorms)

            # Update inv(G)*wm via sherman-morrison
            Ginvwm[:, (m-numpivots):] -= np.outer(Ginvwm[:, m-numpivots+Pind],
                                                  np.dot(A[P[m+Pind], :].T,
                                                  Ginvwm[:, (m-numpivots):])/(1+detnorms[Pind])
                                                  )

            # Pivoting
            P[[m, Pind+m]] = P[[Pind+m, m]]
            Ginvwm[:, [m-numpivots, m-numpivots+Pind]] = Ginvwm[:,
                                                                [m-numpivots+Pind,
                                                                    m-numpivots]]

    return P[:p]


def lstsq_loocv_error(A, b, weights):
    """Computes the leave-one-out cross validation (LOOCV) metric for a
    least-squares problem.

    Parameters:
        A: The M x N design matrix from a least-squares procedure
        b: The right-hand side array with M rows from a least-squares procedure
        weights: size-M array with positive entries, indicating pointwise
            weights in the LOOCV metric.
    Attributes:
        cv: The sum-of-squares cross-validation metric (scalar, float)
    """

    M, N = A.shape
    Q, R, P = qr(A, pivoting=True, mode='economic')

    bdim = len(b.shape)

    if bdim == 1:
        cv = 0.
    else:
        cv = np.zeros(b.shape[1])

    for m in range(M):
        # Leave out m'th sample and compute residal on m'th point.
        Qm, Rm = qr_delete(Q, R, m)
        bm = np.delete(b, m, axis=0)

        if bdim > 1:
            cv += weights[m]*(b[m, :] -
                              A[m, P] @ np.linalg.solve(Rm, Qm.T @ bm))**2
        else:
            cv += weights[m]*(b[m] -
                              A[m, P] @ np.linalg.solve(Rm, Qm.T @ bm))**2

    return np.sqrt(cv/M)


def weighted_lsq(A, b, w):
    """
    Computes the weighted least squares solution to A x = b. I.e., computes x
    that minimizes the :math:`\\ell^2' norm of the weighted residual

    .. math::

      \\mathrm{diag}(\\sqrt{w_1}, \\ldots, \\sqrt{w_M}) (A x - b)


    Parameters:
        A (array-like): The M x N design matrix from a least-squares procedure
        b (array-like): The right-hand side array with M rows from a
            least-squares procedure
        w: size-M array with positive entries, indicating pointwise
            weights in the least squares problem.
    Attributes:
        x (array-like): The least-squares solution.
        residual (array-like): Residual for the least squares problem.
    """

    M, N = A.shape
    assert M == b.shape[0], ("Matrix A and right-hand side b must have "
                             "identical size along axis 0.")

    wsqrt = np.sqrt(w)
    bweighted = np.multiply(b.T, wsqrt).T
    Aweighted = np.multiply(A.T, wsqrt).T

    if version_lessthan(np, '1.14.0'):
        return np.linalg.lstsq(Aweighted, bweighted, rcond=-1)[:2]
    else:
        return np.linalg.lstsq(Aweighted, bweighted, rcond=None)[:2]


if __name__ == "__main__":

    pass



import numpy as np
import scipy as sp
from scipy import sparse as sprs

from UncertainSCI.families import JacobiPolynomials, \
        HermitePolynomials, LaguerrePolynomials
from UncertainSCI.families import DiscreteChebyshevPolynomials
from UncertainSCI.opolynd import TensorialPolynomials
from UncertainSCI.transformations import AffineTransform
from UncertainSCI.utils.version import version_lessthan

# numpy >= 1.17: default_rng is preferred
if version_lessthan(np, '1.17'):
    from numpy.random import choice
else:
    from numpy.random import default_rng
    choice = default_rng().choice


class ProbabilityDistribution:
    def __init__(self):
        pass


class NormalDistribution(ProbabilityDistribution):
    def __init__(self, mean=None, cov=None, dim=None):

        _mean, _cov = self._convert_meancov_to_iterable(mean, cov)

        self._detect_dimension(_mean, _cov, dim)

        assert self.dim > 0, "Dimension must be positive"

        # Construct affine map transformations

        # Low-level routines use Hermite Polynomials
        # with weight function exp(-x**2)
        # instead of standard normal weight function exp(-x**2/2)
        # I.e. x ----> sqrt(2) * x

        A = np.eye(self.dim) * (1/np.sqrt(2))
        b = np.zeros(self.dim)
        self.transform_standard_dist_to_poly = AffineTransform(A=A, b=b)

        # user says: X ~ exp( -(x - mu).T cov^{-1} (x - mu)/2 )
        # want: Z ~ exp( -(x - 0).T eye (x - 0)/2 )
        # I.e. X = KZ + mu, Z = K^{-1}(X - mu), cov = KK.T, K = cov^{1/2}
        # I.e. x ---> cov^{-1/2} * (x + mu) = cov^{-1/2} *x + cov^{-1/2}*mu
        # Opt1: cov = U Lamb U.T, U unitary
        #       cov^{1/2} = U sqrt(Lamb) U.T
        #       cov^{-1/2} = U sqrt(1/Lamb) U.T
        #       A = U sqrt(1/Lamb) U.T and b = A * mu
        # Opt2: cov = L L.T
        #       cov^{1/2} = L
        #       cov^{-1/2} = inv(L)
        #       A = inv(L) and b = A * mu

        if self.dim == 1:
            sigma = np.sqrt(self._cov)
            A = np.eye(self.dim) * (1/sigma)
            b = np.ones(self.dim) * (-self._mean/sigma)
            self.transform_to_standard = AffineTransform(A=A, b=b)

        else:
            # Option 1: Cholesky
            #L = np.linalg.cholesky(self._cov)

            # Option 2: matrix square root
            W, V = np.linalg.eigh(self._cov)
            L = V.T @ np.sqrt(np.diag(W))

            A = np.linalg.inv(L)
            b = -A.dot(self._mean)
            self.transform_to_standard = AffineTransform(A=A, b=b)

        # Construct 1D polynomial families
        Hs = []
        for qd in range(self.dim):
            Hs.append(HermitePolynomials())  # modify for different mean, cov?
        self.polys = TensorialPolynomials(polys1d=Hs)

        # Standard domain is R^dim
        self.standard_domain = np.zeros([2, self.dim])
        self.standard_domain[0, :] = -np.inf
        self.standard_domain[1, :] = np.inf

        self.poly_domain = self.standard_domain.copy()

        self.indices = None

    def _convert_meancov_to_iterable(self, _mean, _cov):
        """
        Converts user-input (mean, cov) to iterables. Ensures that the length
        of the iterables matches on output.

        If mean is None, sets it to 0.
        If cov is None, sets it to identity matrix
        """

        # Tons of type/value checking for mean/cov vs dim
        meaniter = isinstance(_mean, (list, tuple, np.ndarray))

        if _mean is None:
            _mean = 0.

        if _cov is None:
            if meaniter:
                _cov = np.eye(len(_mean))
#            elif _mean is None:
#                _mean = [0.]
#                _cov = np.eye(1)
            else:  # mean is a scalar
                _mean = [_mean, ]
                _cov = np.eye(1)

        else:
            assert isinstance(_cov, np.ndarray), 'Covariance must be an array'
            # assert np.all(cov - cov.T == 0), 'Covariance must be symmetric'

            if meaniter:
                if len(_mean) > 1 and _cov.shape[0] > 1:
                    assert len(_mean) == _cov.shape[0], "Mean and cov parameter \
                            inputs must be of the same size"
                    try:
                        np.linalg.cholesky(_cov)
                    except ValueError:
                        print('Covariance must be symmetric and \
                                positive definite')

#                elif len(_mean) == 1 and _cov.shape[0] == 1:
#                    pass

                elif len(_mean) == 1 and _cov.shape[0] > 1:
                    _mean = [_mean[0] for i in range(_cov.shape[0])]

                elif _cov.shape[0] == 1 and len(_mean) > 1:
                    _cov = np.eye(len(_mean)) * _cov[0]

#            elif _mean is None:
#                _mean = [0. for i in range(_cov.shape[0])]
#
            else:  # mean is a scalar
                _mean = [_mean for i in range(_cov.shape[0])]

        return _mean, _cov

    def _detect_dimension(self, _mean, _cov, dim):
        """
        Parses user-given inputs to determine \
                the dimension of the distribution.

        Mean and cov are iterables.

        Sets self.dim, self.mean, self.cov
        """
        # Situations:
        # 1. User specifies mean list and cov ndarray (disallow contradictory
        #    dimension specification)
        # 2. User specifies dim scalar (disallow contradictory mean, stdev and
        #    cov specification)
        # 3. dim = 1

        if len(_mean) > 1 or _cov.shape[0] > 1:  # Case 1:
            if len(_mean) != _cov.shape[0]:
                raise ValueError('Input parameters mean and cov must \
                        be the same dimension')

            if (dim is not None) and (dim != 1) and (dim != len(_mean)):
                raise ValueError('Mean parameter list must \
                        have size consistent with input dim')

            if (dim is not None) and (dim != 1) and (dim != _cov.shape[0]):
                raise ValueError('Cov parameter array must \
                        have size consistance with input dim')

            self.dim = len(_mean)
            self._mean = _mean
            self._cov = _cov

        elif dim is not None and dim > 1:  # Case 2
            self.dim = dim
            self._mean = [_mean[0] for i in range(self.dim)]
            self._cov = np.eye(self.dim) * _cov[0]

        else:  # Case 3
            self.dim = 1
            self._mean = _mean[0]
            self._cov = _cov[0]

    def MC_samples(self, M=100):
        """
        Returns M Monte Carlo samples from the distribution.
        """
        p = np.random.normal(0., 1., [M, self.dim])

        return self.transform_to_standard.mapinv(p)

    def mean(self):
        return self._mean

    def cov(self):
        return self._cov

    # def means(self):
        # """
        # Returns the mean of the distribution
        # """
#
        # mu = np.zeros(self.dim, )
        # mu = np.reshape(mu, [1, self.dim])
        # return self.transform_to_standard.mapinv(mu).flatten()
#
    # def covariance(self):
        # """
        # Returns the (auto-)covariance matrix of the distribution.
        # """
#
        # sigma = np.eye(self.dim,)
        # zero = np.zeros([1, self.dim])
        # b = self.transform_to_standard.mapinv(zero)
        # bmat = np.tile(b, [self.dim, 1])
        # sigma = self.transform_to_standard.mapinv(sigma) - bmat
        # sigma = (self.transform_to_standard.mapinv(sigma.T) - bmat).T
#
        # return sigma

    def stdev(self):
        """
        Returns the standard deviation of the distribution, if the
        distribution is one-dimensional. Raises an error if called
        for a multivariate distribution.
        """

        if self.dim == 1:
            return np.sqrt(self.cov()[0, 0])
        else:
            raise TypeError("Can only compute standard deviations for scalar\
                             random variables.")

    def pdf(self, x):
        """
        Evaluates the probability density function (pdf) of the distribution at
        the input locations x.
        """
        # density = multivariate_normal.pdf(x, mean=self.mean, cov=self.cov)
        x = self.transform_to_standard.map(x)
        density = np.ones(x.shape[0])
        for i in range(self.dim):
            density *= (2*np.pi)**(-1/2) * np.exp(-1/2 * x[:, i]**2)
        density *= self.transform_to_standard.jacobian_determinant()
        return density


class ExponentialDistribution(ProbabilityDistribution):

    def __init__(self, flag=True, lbd=None, loc=None, mean=None,
                 stdev=None, dim=None):

        # Convert mean/stdev inputs to lbd
        if mean is not None and stdev is not None:
            if lbd is not None:
                raise ValueError('Cannot simultaneously specify lbd \
                        parameter and mean/stdev parameters')

            lbd = self._convert_meanloc_to_lbd(mean, loc)

        lbd, loc = self._convert_lbdloc_to_iterable(lbd, loc)

        # Set self.lbd, self.loc and self.dim
        self._detect_dimension(lbd, loc, dim)

        assert self.dim > 0, "Dimension must be positive"

        # Construct affine map transformations

        # Low-level routines use Laguerre Polynomials, weight
        # x^rho exp^{-x} when rho = 0, which is equal to the standard Beta,
        # exp^{-lbd x} when lbd = 1
        A = np.eye(self.dim)
        b = np.zeros(self.dim)
        self.transform_standard_dist_to_poly = AffineTransform(A=A, b=b)

        if self.dim == 1:
            A = np.array(lbd)
            b = -A * loc
            self.transform_to_standard = AffineTransform(A=A, b=b)

        else:
            A = np.diag(lbd)
            b = -A@loc
            self.transform_to_standard = AffineTransform(A=A, b=b)

        # Construct 1D polynomial families
        Ls = []
        for qd in range(self.dim):
            Ls.append(LaguerrePolynomials())
        self.polys = TensorialPolynomials(polys1d=Ls)

        self.indices = None

    def _detect_dimension(self, lbd, loc, dim):
        """
        Parses user-given inputs to determine the
        dimension of the distribution.

        lbd and loc are iterables.

        Sets self.lbd, self.loc and self.dim
        """

        # Situations:
        # 1. User specifies lbd as a list (disallow contradictory
        #    dimension specification)
        # 2. User specifies lbd as a list, then dim = len(lbd)
        # 3. User specifies dim as a scalar, then lbd = np.ones(dim)
        # 4. User specifies nothing, then dim = 1, lbd = 1

        if len(lbd) > 1 or len(loc) > 1:
            if len(lbd) != len(loc):
                raise ValueError('Input parameters lbd and loc must \
                        be the same dimension')

            if (dim is not None) and (dim != 1) and (dim != len(lbd)):
                raise ValueError('Lbd, loc parameter lists must \
                        have size consistent with input dim')

            self.dim = len(lbd)
            self.lbd = lbd
            self.loc = loc

        elif dim is not None and dim > 1:
            self.dim = dim
            self.lbd = [lbd[0] for i in range(self.dim)]
            self.loc = [loc[0] for i in range(self.dim)]

        else:  # The dimension is 1
            self.dim = 1
            self.lbd = lbd
            self.loc = loc

    def _convert_meanloc_to_lbd(self, mean, loc):
        """
        Converts user-given mean and loc to an iterable lbd.
        """

        meaniter = isinstance(mean, (list, tuple, np.ndarray))
        lociter = isinstance(loc, (list, tuple, np.ndarray))
        lbd = []

        # If they're both iterables:
        if meaniter and lociter:
            assert len(mean) == len(loc)

            for ind in range(len(mean)):
                lb = self.meanloc_to_lbd(mean[ind], loc[ind])
                lbd.append(lb)

        # If mean is an iterable but loc is not
        elif meaniter:
            loc = [loc for i in range(len(mean))]
            for ind in range(len(mean)):
                lb = self.meanloc_to_lbd(mean[ind], loc[ind])
                lbd.append(lb)

        # If loc is an iterable but mean is not
        elif lociter:
            mean = [mean for i in range(len(loc))]
            for ind in range(len(loc)):
                lb = self.meanloc_to_lbd(mean[ind], loc[ind])
                lbd.append(lb)

        # If they're both scalars, let the following lbd checker cf vs dim
        else:
            lbd = self.meanloc_to_lbd(mean, loc)

        return lbd

    def _convert_lbdloc_to_iterable(self, lbd, loc):
        """
        Converts user-input lbd and loc to iterables. Ensures that the length
        of the iterables matches on output.

        If lbd is None, sets it to 1.
        If loc is None, sets it to 0.
        """

        # Tons of type/value checking for lbd/loc vs dim
        lbditer = isinstance(lbd, (list, tuple, np.ndarray))
        lociter = isinstance(loc, (list, tuple, np.ndarray))

        if lbditer and lociter:
            if len(lbd) > 1 and len(loc) > 1:
                assert len(lbd) == len(loc), "Lbd and loc parameter inputs \
                        must be of the same size"
#            elif len(lbd) == 1 and len(loc) == 1:
#                pass
            elif len(lbd) == 1:
                lbd = [lbd[0] for i in range(len(loc))]
            elif len(loc) == 1:
                loc = [loc[0] for i in range(len(lbd))]

        elif lbditer:  # lbd is iterable, loc is not
            if loc is None:
                loc = 0.
            loc = [loc for i in range(len(lbd))]

        elif lociter:  # beta is iterable, alpha is not
            if lbd is None:
                lbd = 1.
            lbd = [lbd for i in range(len(loc))]

        elif lbd is None and loc is None:
            lbd, loc = [1., ], [0., ]

        else:  # alpha, beta should both be scalars
            lbd, loc = [lbd, ], [loc, ]

        return lbd, loc

    def meanloc_to_lbd(self, mu, loc):
        """
        Returns lbd given an input mean (mu) and location (loc)
        for a Exponential distribution
        """

        return 1 / (mu - loc)

    def MC_samples(self, M=100):
        """
        Returns M Monte Carlo samples from the distribution.
        """

        p = np.zeros([M, self.dim])
        for qd in range(self.dim):
            p[:, qd] = np.random.exponential(1, M)

        return self.transform_to_standard.mapinv(p)

    def mean(self):
        mu = np.zeros(self.dim)
        for i in range(self.dim):
            mu[i] = 1.

        mu = np.reshape(mu, [1, self.dim])
        return self.transform_to_standard.mapinv(mu).flatten()

    def cov(self):
        sigma = np.eye(self.dim)
        for i in range(self.dim):
            sigma[i, i] = 1.

        zero = np.zeros([1, self.dim])
        b = self.transform_to_standard.mapinv(zero)
        bmat = np.tile(b, [self.dim, 1])
        sigma = self.transform_to_standard.mapinv(sigma) - bmat
        sigma = (self.transform_to_standard.mapinv(sigma.T) - bmat).T

        return sigma

    def stdev(self):
        """
        Returns the standard deviation of the distribution, if the
        distribution is one-dimensional. Raises an error if called
        for a multivariate distribution.
        """

        if self.dim == 1:
            return np.sqrt(self.cov()[0, 0])
        else:
            raise TypeError("Can only compute standard deviations for \
                    scalar random variables.")

    def pdf(self, x):
        x = self.transform_to_standard.map(x)
        density = np.ones(x.shape[0])
        for i in range(self.dim):
            density *= np.exp(-x[:, i])
        # Scale based on determinant of map
        density *= self.transform_to_standard.jacobian_determinant()

        return density


class GammaDistribution(ProbabilityDistribution):
    """.. _gamma_distribution:

    Constructs a Gamma distribution object. Gamma distributions have support on
    the real interval [0,infty), with probability density function,

    .. math::

      w(y;k,\\theta) := \\frac{1}{\\Gamma(k) \\theta^k}\
              y^{k-1} exp(-y/\\theta), \\hskip 20pt y \\in (0,\\infty),

    where :math:`k` and :math:`\\theta` are positive real parameters that
    define the distribution, and :math:`\\Gamma` is the Gamma function. Some
    special cases of note:

    To generate this distribution on a general shifted interval
    :math:`(shift,\\infty)`, set the shift parameter below.

    To generate this distribution so it has support on the negative half-line
    :math:`(-\\infty,0)`, set the flip parameter below.

    Parameters:
        k (float, optional): Shape parameter. Defaults to 1.
        theta (float, optional): Scale parameter. Defaults to 1.
        shift (float, optional): Shift parameter. Defaults to 0.
        flip (bool, optional): Flip parameter. Defaults to False.

    Attributes:
        k (float): Shape parameter k.
        theta (float): Scale parameter theta.
        polys (:class:`LaguerrePolynomials`): Orthogonal polynomials for this
            distribution.
    """

    def __init__(self, k=1., theta=1., shift=0., flip=False):

        self.theta, self.k, self.shift, self.flip = theta, k, shift, flip
        self.dim = 1

        # Construct affine map transformations

        # Low-level routines use Laguerre Polynomials, weight
        # x^rho exp^{-x} when rho = 0, which is equal to the standard
        # Exponential,
        # exp^{-lbd x} when lbd = 1
        A = np.eye(self.dim)
        b = np.zeros(self.dim)
        self.transform_standard_dist_to_poly = AffineTransform(A=A, b=b)

        if not flip:
            A = 1/self.theta
        else:
            A = -1/self.theta

        b = -A*self.shift
        self.transform_to_standard = AffineTransform(A=A, b=b)

        # Construct 1D polynomial families
        Ls = []
        for qd in range(self.dim):
            Ls.append(LaguerrePolynomials())
        self.polys = TensorialPolynomials(polys1d=Ls)

        self.indices = None

    def MC_samples(self, M=100):
        """
        Returns M Monte Carlo samples from the distribution.
        """

        x = sp.stats.gamma.rvs(self.k, loc=0., scale=self.theta, size=M)
        if self.flip:
            x = -x

        return x + self.shift

    def mean(self):
        mu = self.k
        return self.transform_to_standard.mapinv(mu)

    def stdev(self):
        """
        Returns the standard deviation of the distribution.
        """

        return np.sqrt(self.k)*self.theta

    def pdf(self, x):
        """
        Evaluates the probability distribution function at the locations x.
        """

        z = self.transform_to_standard.map(x)
        density = 1/sp.special.gamma(self.k) * z**(self.k-1) * np.exp(-z)
        return density*self.transform_to_standard.A


class BetaDistribution(ProbabilityDistribution):
    """.. _beta_distribution:

    Constructs a Beta distribution object; supports multivariate distributions
    through tensorization. In one dimension, beta distributions have support on
    the real interval [0,1], with probability density function,

    .. math::

      w(y;\\alpha,\\beta) := \\frac{y^{\\alpha-1} (1-y)^{\\beta-1}}\
              {B(\\alpha,\\beta)}, \\hskip 20pt y \\in (0,1),

    where :math:`\\alpha` and :math:`\\beta` are positive real parameters that
    define the distribution, and :math:`B` is the Beta function. Some special
    cases of note:

    - :math:`\\alpha = \\beta = 1`: the uniform distribution
    - :math:`\\alpha = \\beta = \\frac{1}{2}`: the arcsine distribution

    To generate this distribution on a general compact interval :math:`[a,b]`,
    set the domain parameter below.

    Instead of :math:`(\\alpha, \\beta)`, a mean :math:`\\mu` and standard
    deviation :math:``\\sigma`` may be set. In this case, :math:`(\\mu,
    \\sigma)` must correspond to valid (i.e., positive) values of
    :math:`(\\alpha, \\beta)` on the interval [0,1], or else an error is
    raised.

    Finally, this class supports tensorization: multidimensional distributions
    corresopnding to independent one-dimensional marginal distributions are
    supported. In the case of identically distributed marginals, the `dim`
    parameter can be set to the appropriate dimension. In case of non-identical
    marginals, an array or iterable can be input for :math:`\\alpha, \\beta,
    \\mu, \\sigma`.

    Parameters:
        alpha (float or iterable of floats, optional): Shape parameter
        associated to right-hand boundary. Defaults to 1.
        beta (float or iterable of floats, optional): Shape parameter
        associated to left-hand boundary. Defaults to 1.
        mean (float or iterable of floats, optional): Mean of the distribution.
        Defaults to None.
        stdev (float or iterable of floats, optional): Standard deviation of
        the distribution. Defaults to None.
        dim (int, optional): Dimension of the distribution. Defaults to None.
        domain (numpy.ndarray or similar, of size 2 x `dim`, optional): Compact
        hypercube that is the support of the distribution. Defaults to None

    Attributes:
        dim (int): Dimension of the distribution.
        alpha (float or np.ndarray): Shape parameter(s) alpha.
        beta (float or np.ndarray): Shape parameter(s) beta.
        polys (:class:`JacobiPolynomials` or list thereof):
    """
    def __init__(self, alpha=None, beta=None, mean=None, stdev=None,
                 dim=None, domain=None, bounds=None):

        # Convert mean/stdev inputs to alpha/beta
        if mean is not None and stdev is not None:
            if alpha is not None or beta is not None:
                raise ValueError('Cannot simultaneously specify alpha/beta \
                        parameters and mean/stdev parameters')

            alpha, beta = self._convert_meanstdev_to_alphabeta(mean, stdev)

        alpha, beta = self._convert_alphabeta_to_iterable(alpha, beta)

        if (domain is not None) and (bounds is not None):
            raise ValueError("Inputs 'domain' and 'bounds' cannot both be given")
        elif bounds is not None:
            domain = bounds
        # Sets self.dim, self.alpha, self.beta, self.domain
        self._detect_dimension(alpha, beta, dim, domain)

        for qd in range(self.dim):
            assert self.alpha[qd] > 0 and self.beta[qd] > 0, \
                    "Parameter vectors alpha and beta must have strictly \
                    positive components"
        assert self.dim > 0, "Dimension must be positive"
        assert self.domain.shape == (2, self.dim)

        # Construct affine map transformations

        # Standard domain is [0,1]^dim
        self.standard_domain = np.ones([2, self.dim])
        self.standard_domain[0, :] = 0.

        # Low-level routines use Jacobi Polynomials, which operate on [-1,1]
        # instead of the standard Beta domain of [0,1]
        self.poly_domain = np.ones([2, self.dim])
        self.poly_domain[0, :] = -1.
        self.transform_standard_dist_to_poly = AffineTransform(
                domain=self.standard_domain, image=self.poly_domain)

        self.transform_to_standard = AffineTransform(
                domain=self.domain, image=self.standard_domain)

        # Construct 1D polynomial families
        Js = []
        for qd in range(self.dim):
            Js.append(JacobiPolynomials(alpha=self.beta[qd]-1.,
                      beta=self.alpha[qd]-1.))
        self.polys = TensorialPolynomials(polys1d=Js)

        self.indices = None

    def _detect_dimension(self, alpha, beta, dim, domain):
        """
        Parses user-given inputs to determine
        the dimension of the distribution.

        alpha and beta are iterables.

        Sets self.dim, self.alpha, self.beta, and self.domain
        """

        # Situations:
        # 1. User specifies alpha, beta as lists (disallow contradictory
        #    dimension, domain specification)
        # 2. User specifies dim scalar (disallow contradictory alpha, beta,
        #    domain specification)
        # 3. User specifies domain hypercube

        if len(alpha) > 1 or len(beta) > 1:  # Case 1:
            assert len(alpha) == len(beta), 'Input parameters alpha and beta \
                    must be the same dimension'

            assert (dim is None) or (dim == 1) or (dim == len(alpha)), \
                'Alpha, beta parameter lists must have size consistent with \
                input dim'

            if (domain is not None) and (domain.shape[1] != 1) and \
                    (domain.shape[1] != len(alpha)):
                raise ValueError('Alpha, beta parameter lists must have size \
                        consistent with hypercube domain')

            self.dim = len(alpha)
            self.alpha = alpha
            self.beta = beta

            if domain is None:  # Standard domain [0,1]^dim
                self.domain = np.zeros([2, self.dim])
                self.domain[1, :] = 1.
            else:
                domain = np.asarray(domain)
                if domain.shape[1] == 1:  # Tensorize 1D domain
                    self.domain = np.zeros([2, self.dim])
                    self.domain[0, :] = domain[0]
                    self.domain[1, :] = domain[1]
                else:  # User specified domain
                    self.domain = domain

        elif dim is not None and dim > 1:  # Case 2
            self.dim = dim
            self.alpha = [alpha[0] for i in range(self.dim)]
            self.beta = [beta[0] for i in range(self.dim)]

            if (domain is None) or (domain.any() is not True):  # Standard domain [0,1]^dim
                self.domain = np.zeros([2, self.dim])
                self.domain[1, :] = 1.
            else:
                domain = np.asarray(domain)
                if domain.shape[1] == 1:  # Tensorize 1D domain
                    self.domain = np.zeros([2, self.dim])
                    self.domain[0, :] = domain[0]
                    self.domain[1, :] = domain[1]
                else:  # User specified domain
                    self.domain = domain

            return

        elif domain is not None and len(domain.shape) > 1:  # Case 3
            self.dim = domain.shape[1]
            self.alpha = [alpha[0] for i in range(self.dim)]
            self.beta = [beta[0] for i in range(self.dim)]
            self.domain = domain

        else:  # The dimension is 1
            self.dim = 1
            self.alpha = alpha
            self.beta = beta

            if domain is None:
                self.domain = np.zeros([2, self.dim])
                self.domain[1, :] = 1.
            else:
                self.domain = domain.reshape([2, self.dim])

    def _convert_meanstdev_to_alphabeta(self, mean, stdev):
        """
        Converts user-given mean and stdev to an iterable alpha, beta.
        """

        meaniter = isinstance(mean, (list, tuple))
        stdviter = isinstance(stdev, (list, tuple))
        alpha = []
        beta = []

        # If they're both iterables:
        if meaniter and stdviter:

            # If one has length 1 and the other has length > 1:
            if (len(mean) == 1) and (len(stdev) > 1):
                mean = [mean for i in range(len(stdev))]
            elif (len(stdev) == 1) and (len(mean) > 1):
                stdev = [stdev for i in range(len(mean))]

            for ind in range(len(mean)):
                alph, bet = self.meanstdev_to_alphabeta(mean[ind], stdev[ind])
                alpha.append(alph)
                beta.append(bet)

        # If mean is an iterable but stdev is not
        elif meaniter:
            for ind in range(len(mean)):
                alph, bet = self.meanstdev_to_alphabeta(mean[ind], stdev)
                alpha.append(alph)
                beta.append(bet)

        # If stdev is an iterable but mean is not
        elif stdviter:
            for ind in range(len(mean)):
                alph, bet = self.meanstdev_to_alphabeta(mean, stdev[ind])
                alpha.append(alph)
                beta.append(bet)

        # If they're both scalars, let the following alpha/beta checker
        # cf vs dim
        else:
            alpha, beta = self.meanstdev_to_alphabeta(mean, stdev)

        return alpha, beta

    def _convert_alphabeta_to_iterable(self, alpha, beta):
        """
        Converts user-input (alpha,beta) to iterables. Ensures that the length
        of the iterables matches on output.

        If alpha or beta are None, sets those values to 1.
        """

        # Tons of type/value checking for alpha/beta vs dim
        alphiter = isinstance(alpha, (list, tuple))
        betaiter = isinstance(beta, (list, tuple))
        if alphiter and betaiter:
            if len(alpha) > 1 and len(beta) > 1:
                assert len(alpha) == len(beta), "Alpha and Beta parameter \
                        inputs must be of the same size"
            elif len(alpha) == 1:
                alpha = [alpha[0] for i in range(len(beta))]
            elif len(beta) == 1:
                beta = [beta[0] for i in range(len(alpha))]

        elif alphiter:  # alpha is iterable, beta is not
            if beta is None:
                beta = 1.
            beta = [beta for i in range(len(alpha))]

        elif betaiter:  # beta is iterable, alpha is not
            if alpha is None:
                alpha = 1.
            alpha = [alpha for i in range(len(beta))]

        elif alpha is None and beta is None:
            alpha, beta = [1., ], [1., ]

        else:  # alpha, beta should both be scalars
            alpha, beta = [alpha, ], [beta, ]

        return alpha, beta

    def meanstdev_to_alphabeta(self, mu, stdev):
        """
        Returns alpha, beta given an input mean (mu) and
        standard deviation (stdev) for a Beta distribution
        on the interval [0, 1].
        """

        if 0. >= mu or mu >= 1.:
            raise ValueError('Mean of a standard Beta distribution \
                    must be between 0 and 1.')

        if stdev >= np.sqrt(mu*(1-mu)):
            msg = 'Standard deviation of a Beta random variable must be \
                    smaller than the geometric mean of mu and (1-mu)'
            raise ValueError(msg)

        temp = (mu * (1-mu) - stdev**2)/stdev**2

        return mu*temp, (1-mu)*temp

    def MC_samples(self, M=100):
        """
        Returns M Monte Carlo samples from the distribution.
        """

        p = np.zeros([M, self.dim])
        for qd in range(self.dim):
            p[:, qd] = np.random.beta(self.alpha[qd], self.beta[qd], M)

        return self.transform_to_standard.mapinv(p)

    def mean(self):
        """
        Returns the mean of the distribution.
        """

        mu = np.zeros(self.dim)
        for i in range(self.dim):
            mu[i] = self.alpha[i]/(self.alpha[i] + self.beta[i])

        # Affine map to appropriate interval
        mu = np.reshape(mu, [1, self.dim])
        return self.transform_to_standard.mapinv(mu).flatten()

    def cov(self):
        """
        Returns the (auto-)covariance matrix of the distribution.
        """

        sigma = np.eye(self.dim)
        for i in range(self.dim):
            num = self.alpha[i]*self.beta[i]
            den = (self.alpha[i]+self.beta[i])**2 * \
                  (self.alpha[i]+self.beta[i]+1)
            sigma[i, i] = num/den

        # Need to get scaling part of the affine transform, so need
        # non-homogeneous term b.
        zero = np.zeros([1, self.dim])
        b = self.transform_to_standard.mapinv(zero)
        bmat = np.tile(b, [self.dim, 1])

        # Multiply on right by scaling matrix
        sigma = self.transform_to_standard.mapinv(sigma) - bmat

        # Multiply on left by scaling matrix
        sigma = (self.transform_to_standard.mapinv(sigma.T) - bmat).T

        return sigma

    def stdev(self):
        """
        Returns the standard deviation of the distribution, if the
        distribution is one-dimensional. Raises an error if called
        for a multivariate distribution.
        """

        if self.dim == 1:
            return np.sqrt(self.cov()[0, 0])
        else:
            raise TypeError("Can only compute standard deviations for scalar\
                             random variables.")

    def pdf(self, x):
        """
        Evaluates the probability density function (pdf) of the distribution
        at the input locations x.
        """

        # Evaluate in standard space
        x = self.transform_to_standard.map(x)
        density = np.ones(x.shape[0])
        for i in range(self.dim):
            density *= x[:, i]**(self.alpha[i]-1) * \
                       (1 - x[:, i])**(self.beta[i]-1)
            density /= sp.special.beta(self.alpha[i], self.beta[i])

        # Scale based on determinant of map
        density *= self.transform_to_standard.jacobian_determinant()
        return density


class UniformDistribution(BetaDistribution):
    """.. _uniform_distribution:

    Constructs a (continuous) uniform distribution object; supports
    multivariate distributions through tensorization. In one dimension,
    uniforms distributions have support on the real interval [0,1],
    with probability density function,

    .. math::

      w(y) := 1, \\hskip 20pt y \\in (0,1),

    To generate this distribution on a general compact interval :math:`[a,b]`,
    set the domain parameter below. Alternatively, the mean :math:`\\mu` and
    standard deviation :math:`\\sigma` can be input, in which case the support
    interval :math:`[a,b]` is determined.

    This class supports tensorization: multidimensional distributions
    corresopnding to independent one-dimensional marginal distributions are
    supported. In the case of identically distributed marginals, the `dim`
    parameter can be set to the appropriate dimension. In case of non-identical
    marginals, an array or iterable can be input for :math:`\\alpha, \\beta,
    \\mu, \\sigma`.

    Parameters:
        mean (float or iterable of floats, optional): Mean of the distribution.
        Defaults to None.
        stdev (float or iterable of floats, optional): Standard deviation of
        the distribution. Defaults to None.
        dim (int, optional): Dimension of the distribution. Defaults to None.
        domain (numpy.ndarray or similar, of size 2 x `dim`, optional): Compact
        hypercube that is the support of the distribution. Defaults to None

    Attributes:
        dim (int): Dimension of the distribution.
        polys (:class:`JacobiPolynomials` or list thereof):
    """

    def __init__(self, mean=None, stdev=None, dim=None, domain=None, bounds=None):

        super().__init__(alpha=1., beta=1., mean=mean, stdev=stdev,
                         dim=dim, domain=domain, bounds=bounds)


class DiscreteUniformDistribution(ProbabilityDistribution):
    def __init__(self, n=None, domain=None, dim=None):

        if n is None:
            raise ValueError('Input "n" is required.')

        # Make sure dim is set, and that n is a list with len(n)==dim
        if dim is not None:
            if (len(n) > 1) and (len(n) != dim):
                raise ValueError('Inconsistent settings for inputs "dim" \
                        and "n"')
            elif len(n) == 1:
                if isinstance(n, (list, tuple)):
                    n = [n[0]]*dim
                else:
                    n = [n]*dim

            else:  # len(n) == dim != 1
                pass  # Nothing to do
        else:
            if isinstance(n, (list, tuple)):
                dim = len(n)
            else:
                n = [n]
                dim = 1

        # Ensure that user-specified domain makes sense
        if domain is None:
            domain = np.ones([2, dim])
            domain[0, :] = 0
        else:   # Domain should be a 2 x dim numpy array
            if (domain.shape[0] != 2) or (domain.shape[1] != dim):
                raise ValueError('Inputs "domain" and inferred dimension \
                        are inconsistent')
            else:
                pass  # Nothing to do

        # Assign stuff
        self.dim, self.n, self.domain = dim, n, domain

        # Compute transformations
        # Standard domain is [0,1]^dim
        self.standard_domain = np.ones([2, self.dim])
        self.standard_domain[0, :] = 0.

        # Low-level routines use Discrete Chebyshev Polynomials,
        # which operate on [0,1]
        self.poly_domain = np.ones([2, self.dim])
        self.poly_domain[0, :] = 0.
        self.transform_standard_dist_to_poly = AffineTransform(
                domain=self.standard_domain, image=self.poly_domain)

        self.transform_to_standard = AffineTransform(
                domain=self.domain, image=self.standard_domain)

        Ps = []
        for qd in range(self.dim):
            Ps.append(DiscreteChebyshevPolynomials(M=n[qd]))
        self.polys = TensorialPolynomials(polys1d=Ps)

    def MC_samples(self, M=100):
        """
        Returns M Monte Carlo samples from the distribution.
        """

        p = np.zeros([M, self.dim])
        for qd in range(self.dim):
            p[:, qd] = choice(self.polys.polys1d[qd].standard_support, size=M)

        return self.transform_to_standard.mapinv(p)

    def mean(self):
        mu = np.zeros(self.dim)
        for i in range(self.dim):
            mu[i] = 1/2
        mu = np.reshape(mu, [1, self.dim])
        return self.transform_to_standard.mapinv(mu).flatten()

    def cov(self):
        sigma = np.eye(self.dim)
        for i in range(self.dim):
            sigma[i, i] = (self.n[i]+1) / (12*(self.n[i]-1))
        zero = np.zeros([1, self.dim])
        b = self.transform_to_standard.mapinv(zero)
        bmat = np.tile(b, [self.dim, 1])
        sigma = self.transform_to_standard.mapinv(sigma) - bmat
        sigma = (self.transform_to_standard.mapinv(sigma.T) - bmat).T

        return sigma

    def stdev(self):
        """
        Returns the standard deviation of the distribution, if the distribution
        is one-dimensional. Raises an error if called for a multivariate
        distribution.
        """

        if self.dim == 1:
            return np.sqrt(self.cov()[0, 0])
        else:
            raise TypeError("Can only compute standard deviations for scalar\
                             random variables.")

    def pmf(self, x):
        """
        Evaluates the probability mass function (pmf) of the distribution
        """
        density = np.ones(x.shape[0])
        for i in range(self.dim):
            density *= 1/self.n[i]
        return density


class TensorialDistribution(ProbabilityDistribution):
    def __init__(self, distributions=None, dim=None):

        if dim is not None:
            if len(distributions) > 1:
                raise ValueError("Input 'dim' cannot be set if \
                        'distributions' contains more than one element")
            else:
                distributions = dim*distributions

        self.distributions = distributions
        self.dim = np.sum([dist.dim for dist in distributions])

        self.standard_domain = np.concatenate(
                [dist.standard_domain.T for dist in distributions]
                ).T

        self.poly_domain = np.concatenate(
                [dist.poly_domain.T for dist in distributions]
                ).T

        # Construct transform_standard_dist_to_poly
        Ts = [dist.transform_standard_dist_to_poly for dist in distributions]
        As = [t.A.toarray() if isinstance(t.A, sprs.spmatrix) else
              t.A for t in Ts]
        bs = [t.b for t in Ts]

        self.transform_standard_dist_to_poly = AffineTransform(
                A=sp.linalg.block_diag(*As),
                b=np.concatenate(bs)
            )

        # Construct transform_to_standard
        Ts = [dist.transform_to_standard for dist in distributions]
        As = [t.A.toarray() if isinstance(t.A, sprs.spmatrix) else
              t.A for t in Ts]
        bs = [t.b for t in Ts]

        self.transform_to_standard = AffineTransform(
                A=sp.linalg.block_diag(*As),
                b=np.concatenate(bs)
            )

        self.polys = TensorialPolynomials([poly for dist in distributions
                                          for poly in dist.polys.polys1d])

        self.indices = None

    def MC_samples(self, M=100):
        """
        Returns M Monte Carlo samples from the distribution
        """

        p = np.zeros([M, self.dim])
        counter = 0
        for dist in self.distributions:
            p[:, range(counter, counter+dist.dim)] = dist.MC_samples(M=M)
            counter += dist.dim

        # Each component distribution already applies
        # transform_to_standard.mapinv.
        return p



# Contains some rudimentary (physical-space) models for testing PCE
# approximations. All these functions support the syntax output = f(p), where p
# is a d-dimensional vector, and output is a vector whose size is the dimension
# of the model output.

import numpy as np
from scipy import sparse
from scipy.sparse import linalg as splinalg
import scipy.optimize

def ishigami_function(a, b):
    """
    Returns the Ishigami function,

    .. math::

      f(p) = (1 + b*p_3^4) \\sin p_1 + a \\sin^2 p_2,

    where :math:`p = (p_1, p_2, p_3)` are random parameters that are all
    uniformly distributed over :math:`[-\\pi, \\pi]`: :math:`p_i \\sim
    \\mathcal{U}([-\\pi, \\pi])`.
    """

    def f(p):
        return np.sin(p[0]) * (1 + b*p[2]**4) + a*np.sin(p[1])**2

    return f

def borehole_function():
    """
    Returns the Borehole function,

    .. math::

      f(p) = g_1(p) / (g_2(p) * g_3(p)),
      g_1(p) = 2\\pi p_3 * (p_4 - p_6)
      g_2(p) = log(p_2/p_1)
      g_3(p) = 1 + 2*p_7*p_3/(g_2(p) * p_1^2 * p_8) + p_3/p_5

    where the 8-dimensional parameters :math:`p` are

      p = (p_1, p_2, p_3, p_4, p_5, p_6, p_7, p_8)
        = (r_w, r,   T_u, H_u, T_l, H_l, L,   K_w)
    """

    def g1(p):
        return 2*np.pi*p[2]*(p[3] - p[5])

    def g2(p):
        return np.log(p[1]/p[0])

    def g3(p, g2val):
        return 1 + 2*p[6]*p[2]/(g2val * p[0]**2 * p[7]) + p[2]/p[4]

    def f(p):
        g2val = g2(p)
        return g1(p)/(g2val * g3(p,g2val))

    return f

def taylor_frequency(p):
    """
    Returns ( \\sum_{j=1}^d p_j^j )
    """

    return np.sum(p**(1 + np.arange(p.size)))


def sine_modulation(left=-1, right=1, N=100):
    """
    For a d-dimensional parameter p, defines the model,

      f(x,p) = sin [ pi * ( \\sum_{j=1}^d p_j^j ) * x ],

    where x is N equispaced points on the interval [left, right].

    Returns a function pointer with the syntax p ----> f(p).
    """

    x = np.linspace(left, right, N)

    return lambda p: np.sin(np.pi * x * taylor_frequency(p))


def mercer_eigenvalues_exponential_kernel(N, a, b):
    """
    For a 1D exponential covariance kernel,

      K(s,t) = exp(-|t-s| / a),     s, t \\in [-b,b],

    computes the first N eigenvalues of the associated Mercer integral
    operator.

    Precisely, computes the first N/2 positive solutions to both of the following
    transcendental equations for w and v:

       1 - a v tan(v b) = 0
       a w + tan(w b) = 0

    The eigenvalues are subsequently defined through these solutions.

    Returns (1) the N eigenvalues lamb, (2) the first ceil(N/2) solutions for
    v, (3) the first floor(N/2) solutions for w.
    """

    assert N > 0 and a > 0 and b > 0

    M = int(np.ceil(N/2))
    w = np.zeros(M)
    v = np.zeros(M)

    # First equation transformed:
    # vt = v b
    #
    #  -(b/a) / vt + tan(vt) = 0

    def f(x):
        return -(b/a)/x + np.tan(x)

    for n in range(M):
        # Compute bracketing interval
        # root somewhere in right-hand part of [2*n-1, 2*n+1]*pi/2 interval
        RH_value = -1
        k = 4
        while RH_value < 0:
            k += 1
            right = (2*n+1)*np.pi/2 - 1/k
            RH_value = f(right)

        # Root can't be on LHS of interval
        if n == 0:
            left = 1/k
            while f(left) > 0:
                k += 1
                left = 1/k
        else:
            left = n*np.pi

        v[n] = scipy.optimize.brentq(f, left, right)

    v /= b

    # Second equation transformed:
    # wt = w b
    #
    #  (a/b) wt + tan(wt) = 0

    def f(x):
        return (a/b)*x + np.tan(x)

    for n in range(M):

        # Compute bracketing interval
        # root somewhere in [2*n+1, 2*n+3]*pi/2
        LH_value = 1
        k = 4
        while LH_value > 0:
            k += 1
            left = (2*n+1)*np.pi/2 + 1/k
            LH_value = f(left)

        # Root can't be on RHS of interval
        right = (n+1)*np.pi

        w[n] = scipy.optimize.brentq(f, left, right)

    w /= b

    if (N % 2) == 1:  # Don't need last root for w
        w = w[:-1]

    lamb = np.zeros(N)
    oddinds = [i for i in range(N) if (i % 2) == 0]  # Well, odd for 1-based indexing
    lamb[oddinds] = 2*a/(1+(a*v)**2)
    eveninds = [i for i in range(N) if (i % 2) == 1]  # even for 1-based indexing
    lamb[eveninds] = 2*a/(1+(a*w)**2)

    return lamb, v, w


def KLE_exponential_covariance_1d(N, a, b, mn):
    """
    Returns a pointer to a function the evaluates an N-term Karhunen-Loeve
    Expansion of a stochastic process with exponential covariance function on a
    bounded interval [-b,b]. Let the GP have the covariance function,

        C(s,t) = exp(-|t-s| / a),

    and mean function given by mn. Then the N-term KLE of the process is given
    by

        K_N(x,P) = mn(x) + \\sum_{n=1}^N P_n sqrt(\\lambda_n) \\phi_n(x),

    where (lambda_n, phi_n) are the leading eigenpairs of the associated Mercer
    kernel. The eigenvalues are computed in
    mercer_eigenvalues_exponential_kernel. The (P_n) are iid standard normal
    Gaussian random variables.

    Returns a function lamb(x,P) that takes in a 1D np.ndarray x and a 1D
    np.ndarray vector P and returns the KLE realization on x for that value of
    P.
    """

    lamb, v, w = mercer_eigenvalues_exponential_kernel(N, a, b)

    efuns = N*[None]
    for i in range(N):
        if (i % 2) == 0:
            i2 = int(i/2)
            efuns[i] = (lambda i2: lambda x: np.cos(v[i2]*x) / np.sqrt(b + np.sin(2*v[i2]*b)/(2*v[i2])))(i2)
        else:
            i2 = int((i-1)/2)
            efuns[i] = (lambda i2: lambda x: np.sin(w[i2]*x) / np.sqrt(b - np.sin(2*w[i2]*b)/(2*w[i2])))(i2)

    def KLE(x, p):
        return mn(x) + np.array([np.sqrt(lamb[i])*efuns[i](x) for i in range(N)]).T @ p

    return KLE


def laplace_ode_diffusion(x, p):
    """ Parameterized diffusion coefficient for 1D ODE

    For a d-dimensional parameter p, the diffusion coefficient a(x,p) has the form

      a(x,p) = pi^2/5 + sum_{j=1}^d p_j * sin(j*pi*(x+1)/2) / j^2,

    which is positive for all x if all values of p lie between [-1,1].
    """

    a_val = np.ones(x.shape)*np.pi**2/5
    for q in range(p.size):
        a_val += p[q] * np.sin((q+1)*np.pi*(x+1)/2)/(q+1)**2
    return a_val


def laplace_grid_x(left, right, N):
    """
    Computes one-dimensional equispaced grid with N points on the interval
    (left, right).
    """
    return np.linspace(left, right, N)


def laplace_ode(left=-1., right=1., N=100, f=None, diffusion=laplace_ode_diffusion):
    """

    Computes the solution to the ODE:

      -d/dx [ a(x,p) d/dx u(x,p) ] = f(x),

    with homogeneous Dirichlet boundary conditions at x = left, x = right.

    For a d-dimensional parameter p, a(x,p) is the function defined in laplace_ode_diffusion.

    Uses an equispaced finite-difference discretization of the ODE.

    """

    assert N > 2

    if f is None:
        def f(x):
            return np.pi**2 * np.cos(np.pi*x)

    x = laplace_grid_x(left, right, N)
    h = x[1] - x[0]
    fx = f(x)

    # Set homogeneous Dirichlet conditions
    fx[0], fx[-1] = 0., 0.

    # i+1/2 points
    xh = x[:-1] + h/2.

    def create_system(p):
        nonlocal x, xh, N
        a = diffusion(xh, p)
        number_nonzeros = 1 + 1 + (N-2)*3
        rows = np.zeros(number_nonzeros, dtype=int)
        cols = np.zeros(number_nonzeros, dtype=int)
        vals = np.zeros(number_nonzeros, dtype=float)

        # Set the homogeneous Dirichlet conditions
        rows[0], cols[0], vals[0] = 0, 0, 1.
        rows[1], cols[1], vals[1] = N-1, N-1, 1.
        ind = 2

        for q in range(1, N-1):
            # Column q-1
            rows[ind], cols[ind], vals[ind] = q, q-1, -a[q-1]
            ind += 1

            # Column q
            rows[ind], cols[ind], vals[ind] = q, q, a[q-1] + a[q]
            ind += 1

            # Column q+1
            rows[ind], cols[ind], vals[ind] = q, q+1, -a[q]
            ind += 1

        A = sparse.csc_matrix((vals, (rows, cols)), shape=(N, N))

        return A

    def solve_system(p):
        nonlocal fx, h
        return splinalg.spsolve(create_system(p), fx*(h**2))

    return lambda p: solve_system(p)

def laplace_ode_1d(Nparams, a=1., b=1., abar=3., N=100):
    """ Generates a 1D Laplace ODE model with parameterized diffusion.

    Define model:

    -d/dx a(x,p) d/dx u(x,p) = f(x)

    over x in [-1,1], where a(x,p) is a parameterized diffusion model:

    a(x,p) = abar(x) + sum_{j=1}^d lambda_j Y_j phi_j(x),

    where d = Nparams, (lambda_j, phi_j) are eigenpairs of the exponential
    covariance kernel,

      K(s,t) = exp(-|s-t|/a).

    The Y_j are modeled as iid random variables.
    """

    abarfun = lambda x: abar*np.ones(np.shape(x))
    KLE = KLE_exponential_covariance_1d(Nparams, a, b, abarfun)

    diffusion = lambda x, p: KLE(x, p)
    x = laplace_grid_x(-b, b, N)

    return x, laplace_ode(left=-b, right=b, N=N, diffusion=diffusion)

def laplace_grid_xy(left, right, N1, down, up, N2):
    """
    Computes two-dimensional tensorial equispaced grid corresponding to the
    tensorization of N1 equispaced points on the interval (left, right) and N2
    equispaced points on the interval (down, up).
    """
    x = np.linspace(left, right, N1)
    y = np.linspace(down, up, N2)

    X, Y = np.meshgrid(x, y)

    return X.flatten(order='C'), Y.flatten(order='C')


def laplace_pde_diffusion(x, p):
    """ Parameterized diffusion coefficient for 2D PDE

    For a d-dimensional parameter p, the diffusion coefficient a(x,p) has the form

      a(x,p) = pi^2/5 + sum_{j=1}^d p_j * sin(j*pi*(x+1)/2) / j^2,

    which is positive for all x if all values of p lie between [-1,1].
    """

    a_val = np.ones(x.shape)*np.pi**2/5
    for q in range(p.size):
        a_val += p[q] * np.sin((q+1)*np.pi*(x+1)/2)/(q+1)**2
    return a_val


def genz_oscillatory(w=0., c=None):
    """
    Returns a pointer to the "oscillatory" Genz test function defined as

       f(p) = \\cos{ 2\\pi w + \\sum_{i=1}^dim c_i p_i }

    where p \\in R^d. The default value for w is 0, and that for c is a
    d-dimensional vector of ones.
    """

    def cos_eval(p):
        nonlocal c
        if c is None:
            c = np.ones(p.size)
        return np.cos(2*np.pi*w + np.dot(c, p))

    return lambda p: cos_eval(p)


if __name__ == "__main__":

    from matplotlib import pyplot as plt
    import scipy as sp

    dim = 5

    a = 3
    b = 1

    def mn(x):
        return np.zeros(x.shape)

    KLE = KLE_exponential_covariance_1d(dim, a, b, mn)

    def diffusion(x, p):
        return np.exp(KLE(x, p))

    left = -1.
    right = 1.
    N = 1000
    model = laplace_ode(left=left, right=right, N=N, diffusion=diffusion)
    x = laplace_grid_x(left, right, N)

    K = 4
    p = K*[None]
    u = K*[None]
    a = K*[None]

    for k in range(K):
        p[k] = np.random.rand(dim)*2 - 1
        # a[k] = laplace_ode_diffusion(x, p[k])
        a[k] = diffusion(x, p[k])
        u[k] = model(p[k])

    for k in range(K):

        row = np.floor(k/2) + 1
        col = k % (K/2) + 1

        index = col + (row-1)*K/2

        plt.subplot(2, K, k+1)
        plt.plot(x, a[k], 'r')
        plt.title('Diffusion coefficient')
        plt.ylim([0, 3.0])

        plt.subplot(2, K, k+1+K)
        plt.plot(x, u[k])
        plt.title('Solution u')
        plt.ylim([-5, 5])

    M = 1000
    U = np.zeros([u[0].size, M])
    for m in range(M):
        U[m, :] = model(np.random.rand(dim)*2 - 1)

    _, svs, _ = np.linalg.svd(U)
    _, r, _ = sp.linalg.qr(U, pivoting=True)

    plt.figure()
    plt.semilogy(svs[:100], 'r')
    plt.semilogy(np.abs(np.diag(r)[:100]), 'b')
    plt.legend(["Singular values", "Orthogonalization residuals"])
    plt.show()



import numpy as np
import scipy.special as sp

# from UncertainSCI.utils.compute_subintervals import compute_subintervals
from UncertainSCI.utils.quad import compute_subintervals,\
                                    gq_modification_composite
from UncertainSCI.utils.quad import gq_modification_unbounded_composite


def compute_moment_bounded(a, b, weight, n, singularity_list, Nquad=10):
    """
    Return the first 2n+1 (finite) moments, {mu_i}_{i=0}^2n
    """
    assert a < b

    subintervals = compute_subintervals(a, b, singularity_list)

    m = np.zeros(2*n+1,)
    for i in range(len(m)):
        def integrand(x):
            return weight(x) * x**i
        # integrand = lambda x: weight(x) * x**i
        m[i] = gq_modification_composite(integrand, a, b, i+1+Nquad,
                                         subintervals)

    return m


def compute_moment_unbounded(a, b, weight, n, singularity_list, Nquad=10):

    assert a < b

    m = np.zeros(2*n+1,)
    for i in range(len(m)):
        def integrand(x):
            return weight(x) * x**i
        # integrand = lambda x: weight(x) * x**i
        m[i] = gq_modification_unbounded_composite(integrand, a, b, i+1+Nquad,
                                                   singularity_list)

    return m


def compute_moment_discrete(xg, wg, n):
    m = np.zeros(2*n+1,)
    for i in range(len(m)):
        # integrand = lambda x: x**i
        def integrand(x):
            return x**i
        m[i] = np.sum(integrand(xg) * wg)

    return m


def compute_freud_moment(rho, m, n):
    """
    with respect to the Freud weight w_rho(x) by gamma function,
    w_rho(x) = |x|^rho * exp(-|x|^m), rho > -1, m > 0.
    """
    mom = np.zeros(2*n+1)
    for i in range(2*n+1):
        if i % 2 == 0:
            mom[i] = 2 * sp.gamma((i+1+rho)/m) / m
        else:
            mom[i] = 0
    return mom


if __name__ == '__main__':
    m1 = compute_freud_moment(rho=0, m=2, n=5)
    m2 = compute_moment_unbounded(a=-np.inf, b=np.inf,
                                  weight=lambda x: np.exp(-x**2),
                                  n=5, singularity_list=[])
    print(np.linalg.norm(m1 - m2, None))



import numpy as np
from matplotlib import pyplot as plt

def piechart_sensitivity(pce, ax=None, scalarization=None, interaction_orders=None, **kwargs):
    """Constructs and plots a pie chart of global sensitivities.
    """

    pce.assert_pce_built()

    if ax is None:
        ax = plt.figure().gca()

    if scalarization is None:
        def scalarization(GI):
            # Default scalarization: mean value
            return np.mean(GI, axis=1)

    global_sensitivity, variable_interactions = pce.global_sensitivity(interaction_orders=interaction_orders)
    scalarized_GSI = scalarization(global_sensitivity)

    labels = [' '.join([pce.plabels[v] for v in varlist]) for varlist in variable_interactions]

    ax.pie(scalarized_GSI*100, labels=labels, autopct='%1.1f%%', startangle=90)

    ax.axis('equal')  # Equal aspect ratio ensures that pie is drawn as a circle.
    plt.title('Sensitivity due to variable interactions')

    ax.set(**kwargs)

    return ax

def quantile_plot(pce, bands=3, xvals=None, ax=None, datadim=1, **kwargs):
    """Plots quantiles.

    bands: number of quantile bands around the median.
    """

    pce.assert_pce_built()

    if datadim != 1:
        raise NotImplementedError("Visualization is only implemented for 1D data")

    if xvals is None:
        x = np.arange(pce.coefficients.shape[1])
    else:
        x = np.asarray(xvals)
        assert x.size == pce.coefficients.shape[1]

    dq = 0.5/(bands+1)
    q_lower = np.arange(dq, 0.5-1e-7, dq)[::-1]
    q_upper = np.arange(0.5 + dq, 1.0-1e-7, dq)
    quantile_levels = np.append(np.concatenate((q_lower, q_upper)), 0.5)

    quantiles = pce.quantile(quantile_levels, M=int(2e3))
    median = quantiles[-1, :]

    band_mass = 1/(2*(bands+1))

    # quantile plot
    if ax is None:
        ax = plt.figure().gca()

    ax.plot(x, median, 'b', label='Median')

    for ind in range(bands):
        alpha = (bands-ind) * 1/bands - (1/(2*bands))
        if ind == 0:
            ax.fill_between(x, quantiles[ind, :], quantiles[bands+ind, :],
                             interpolate=True, facecolor='red', alpha=alpha,
                             label='{0:1.2f} probability mass (each band)'.format(band_mass))
        else:
            ax.fill_between(x, quantiles[ind, :], quantiles[bands+ind, :], interpolate=True, facecolor='red', alpha=alpha)

    ax.set_title('Median + quantile bands')
    ax.set_xlabel('$x$')
    ax.legend(loc='lower right')

    ax.set(**kwargs)

    return ax

def mean_stdev_plot(pce, ensemble=False, xvals=None, ax=None, datadim=1, **kwargs):
    """Plots mean, standard deviation, with optional ensemble overlay.
    """

    pce.assert_pce_built()

    if datadim != 1:
        raise NotImplementedError("Visualization is only implemented for 1D data")

    if xvals is None:
        x = np.arange(pce.coefficients.shape[1])
    else:
        x = np.asarray(xvals)
        assert x.size == pce.coefficients.shape[1]

    if ensemble:
        output = np.zeros([ensemble, pce.coefficients.shape[1]])
        p = pce.distribution.MC_samples(ensemble)
        for j in range(ensemble):
            output[j,:] = pce.eval(p[j,:])

    if ax is None:
        ax = plt.figure().gca()

    mean = pce.mean()
    stdev = pce.stdev()

    # mean +/- stdev plot
    if ensemble:
        ax.plot(x, output[:ensemble, :].T, 'k', alpha=0.8, linewidth=0.2)
    ax.plot(x, mean, 'b', label='Mean')
    ax.fill_between(x, mean-stdev, mean+stdev, interpolate=True, facecolor='red', alpha=0.5, label='$\pm 1$ stdev range')

    ax.set_xlabel('$x$')
    ax.set_title('Mean $\\pm$ standard deviation')

    ax.legend(loc='lower right')

    ax.set(**kwargs)

    return ax



import numpy as np
from UncertainSCI import opoly1d
from UncertainSCI.opoly1d import OrthogonalPolynomialBasis1D

from UncertainSCI.utils.linalg import greedy_d_optimal
from UncertainSCI.sampling import mixture_tensor_discrete_sampling

def opolynd_eval(x, lambdas, ab):
    # Evaluates tensorial orthonormal polynomials associated with the
    # univariate recurrence coefficients ab.

    try:
        M, d = x.shape
    except Exception:
        d = x.size
        M = 1
        x = np.reshape(x, (M, d))

    N, d2 = lambdas.shape

    assert d == d2, "Dimension 1 of x and lambdas must be equal"

    p = np.ones([M, N])

    for qd in range(d):
        p = p * opoly1d.eval_driver(x[:, qd], lambdas[:, qd], 0, ab)

    return p


class TensorialPolynomials:
    def __init__(self, polys1d=None, dim=None):

        if polys1d is None:
            raise TypeError('A one-dimemsional polynomial family is required.')

        # Isotropic tensorization of 1D family
        elif isinstance(polys1d, OrthogonalPolynomialBasis1D):
            self.polys1d = polys1d
            self.isotropic = True
            if dim is None:
                self.dim = 1
            else:
                self.dim = dim

        # Anistropic tensorization of 1D family
        elif isinstance(polys1d, list) or isinstance(polys1d, tuple):
            if dim is not None:
                raise ValueError('If polys1d is a list/tuple, the scalar dim \
                                  cannot be set.')
            for item in polys1d:
                if not isinstance(item, OrthogonalPolynomialBasis1D):
                    raise ValueError('Elements of polys1d must be \
                                      OrthogonalPolynomialBasis1D objects')
            self.polys1d = list(polys1d)
            self.isotropic = False
            self.dim = len(self.polys1d)

        else:
            raise TypeError('Unrecognized type for input polys1d')

    def eval(self, x, lambdas, d=None):
        """
        Evaluates tensorial orthonormal polynomials.
        """

        try:
            M, dim = x.shape
        except Exception:
            dim = x.size
            M = 1
            x = np.reshape(x, (M, dim))

        N, dim2 = lambdas.shape

        assert dim == dim2 == self.dim, "Dimension 0 of both x and lambdas must \
                                         equal self.dim"

        if d is None:
            d = [0, ]*dim

        assert len(d) == dim, 'Derivative order must be a list-type ' \
                              'with length equal to dimension.'

        p = np.ones([M, N])

        if self.isotropic:
            for qd in range(self.dim):
                p = p * self.polys1d.eval(x[:, qd], lambdas[:, qd], d=d[qd])
        else:
            for qd in range(self.dim):
                p = p * self.polys1d[qd].eval(x[:, qd], lambdas[:, qd], d=d[qd])

        return p

    def idist_mixture_sampling(self, M, Lambdas, weights=None,
                               fast_sampler=True):
        """
        Performs tensorial inverse transform sampling from an additive mixture
        of tensorial induced distributions, generating M samples

        The measure this samples from is the order-Lambdas induced measure,
        which is an additive mixture of tensorial measures

        Each tensorial measure is defined a row of Lambdas

        Parameters
        ------
        param1: M
        Number of samples to generate
        param2: Lambdas
        Sample from the order-Lambdas induced measure

        Returns
        ------
        """

        from UncertainSCI.utils.prob import discrete_sampling

        K, d = Lambdas.shape

        assert M > 0 and d == self.dim

        if weights is None:
            weights = np.ones(K)/K
        else:
            assert weights.size == K
            assert np.all(weights >= 0)
            weights = weights/np.sum(weights)

        ks = discrete_sampling(M, weights, np.arange(K, dtype=int))
        Lambdas = Lambdas[ks, :]

        if self.isotropic:

            fidistinv = getattr(self.polys1d, "fidistinv", None)
            if callable(fidistinv):
                has_fidistinv = True
            else:
                has_fidistinv = False

            if fast_sampler and has_fidistinv:
                idistinv = self.polys1d.fidistinv
            else:
                idistinv = self.polys1d.idistinv

            return idistinv(np.random.random([M, d]), Lambdas)

        else:
            x = np.zeros([M, d])
            for qd in range(self.dim):

                fidistinv = getattr(self.polys1d[qd], "fidistinv", None)
                if callable(fidistinv):
                    has_fidistinv = True
                else:
                    has_fidistinv = False

                if fast_sampler and has_fidistinv:
                    idistinv = self.polys1d[qd].fidistinv
                else:
                    idistinv = self.polys1d[qd].idistinv

                x[:, qd] = idistinv(np.random.random(M), Lambdas[:, qd])
            return x

    def idist_gq_sampling(self, K, iset, M=None):
        """
        Generates K random samples from the induced distribution defined
        by Lambdas, restricted to the tensorial Gauss quadrature grid
        defined by M.
        """

        max_degrees = iset.max_univariate_degree()

        if M is None:
            M = [m+1 for m in max_degrees]
            self.sampling_options['M'] = M
        else:
            try:
                iter(M)
            except:
                M = [M,]*self.dim

        # Construct all tensorial measures
        grids = [None,]*self.dim
        weights = [None,]*self.dim
        for j in range(self.dim):
            # Gauss grid
            grids[j], w = self.polys.polys1d[j].gauss_quadrature(M[j])

            # Measures
            kj = max_degrees[j]+1
            weights[j] = np.zeros([grids[j].size, kj])
            V = self.polys1d[j].eval(grids[j], range(max_degrees[j]+1))**2
            V = V * (np.tile(w, [kj, 1]).T)

            for k in range(kj):
                # Ensure probabilities sum to 1
                weights[:,k] = V[:,k]/np.sum(V[:,k])

        return mixture_tensor_discrete_sampling(K, grids, weights, iset)

    def wafp_sampling(self, indices, oversampling=10, weights=None,
                      sampler='idist', K=None, fast_sampler=True):
        """
        Computes (indices.shape[0] + oversampling) points using the WAFP
        strategy. This requires forming K random samples; the input sampler
        determines which measure to take these random samples from.
        """

        M = indices.shape[0] + oversampling

        if K is None:
            K = M + 40
        else:
            K = max(K, M)

        if sampler == 'idist':
            x = self.idist_mixture_sampling(K, indices, weights=weights,
                                            fast_sampler=fast_sampler)
        else:
            raise ValueError('Unrecognized string "{0:s}" for input \
                             "sampler"'.format(sampler))

        V = self.eval(x, indices)

        # Precondition rows to have unit norm
        V = np.multiply(V.T, 1/np.sqrt(np.sum(V**2, axis=1))).T

        P = greedy_d_optimal(V, M)

        return x[P, :]

    def wafp_sampling_restart(self, indices, samples, Nnewsamples,
                              weights=None, sampler='idist',
                              fast_sampler=True):
        """
        Computes Nnewsamples from a "restarted" WAFP procedure, from which the
        input samples are already prescribed.
        """

        if Nnewsamples == 0:
            return samples

        Mold = samples.shape[0]
        V = self.eval(samples, indices)
        # Precondition rows to have unit norm
        V = np.multiply(V.T, 1/np.sqrt(np.sum(V**2, axis=1))).T

        # We'll do this slowly since adaptivity is hard.
        # Generate Ncandidate new candidates, choose 1 new sample, repeat.
        Ncandidates = 100

        for q in range(Nnewsamples):

            if sampler == 'idist':
                x = self.idist_mixture_sampling(Ncandidates, indices,
                                                weights=weights,
                                                fast_sampler=fast_sampler)
            else:
                raise ValueError('Unrecognized string "{0:s}" for input \
                                  "sampler"'.format(sampler))

            temp = self.eval(x, indices)
            temp = np.multiply(temp.T, 1/np.sqrt(np.sum(temp**2, axis=1))).T

            A = np.vstack((V, temp))
            P = greedy_d_optimal(A, Mold+q+1, pstart=range(Mold+q))

            # Append the new sample to both V and samples
            V = np.vstack((V, A[P[-1], :]))
            samples = np.vstack((samples, x[P[-1]-Mold-q, :]))

        return samples

    def tensor_gauss_quadrature(self, N):
        """ Tensorial Gaussian quadrature

        Generates a self.dim-dimensional tensorial rule constructed from the
        univariate Gaussian quadrature rules from each polynomial family.

        Args:
            N: If given as an integer, specifies an equal number of points in
              each dimension. If a size-len(self.dim) iterable of integers,
              specifies the number of points in each separate dimension.
        Returns:
            x: A shape-(K x self.dim) numpy array, where each row corresponds
              to a quadrature node.
            w: A size-K numpy array containing corresponding quadrature
              weights.
        """

        try:
            iter(N)
        except Exception:
            N = [N, ]*self.dim

        xs = [None, ]*self.dim
        ws = [None, ]*self.dim

        for q in range(self.dim):
            if self.isotropic:
                xs[q], ws[q] = self.polys1d.gauss_quadrature(N[q])
            else:
                xs[q], ws[q] = self.polys1d[q].gauss_quadrature(N[q])

        X = np.meshgrid(*xs)
        W = np.meshgrid(*ws)

        x = np.zeros([X[0].size, self.dim])
        w = np.ones(X[0].size)

        for q in range(self.dim):
            x[:, q] = X[q].flatten()
            w *= W[q].flatten()

        return x, w

    def get_univariate_recurrence(self, N, dim):
        """ Returns univariate three-term recurrence coefficient

        Returns the first N+1 recurrence coefficients for the family in
        dimension dim.

        Args:
            N: Integer, determining the maximum index of the coefficients
                desired.
            dim: Integer, the dimension index, satisfying 0 <= dim < self.dim
        Returns:
            ab: A shape-(N+1 x 2) numpy array containing the coefficients.
        """

        assert dim < self.dim, 'Requested dimension index larger' \
                               'than allowed.'

        if self.isotropic:
            return self.polys1d.recurrence(N)
        else:
            return self.polys1d[dim].recurrence(N)

    def get_univariate_derivative_expansion(self, dim, s, N, K=None):
        """ Returns univariate derivative expansion coefficients

        Returns the (N+1) x (K+1) s-th derivative expansion array associated to
        the given dimension dim.

        Args:
            dim: Integer, the dimension index, satisfying 0 <= dim < self.dim
            s: The integer order of the derivative. Must be non-negative.
            N: Computes coefficients for :math:`n \\leq N`
            K: Computes coefficients for :math:`k \\leq K` (optional). If not
              given, is set to N.
        Returns:
            C: (N+1) x (K+1) numpy array containing coefficients.
        """
        assert dim < self.dim, 'Requested dimension index larger' \
                               'than allowed.'

        if self.isotropic:
            return self.polys1d.derivative_expansion(s, N, K=K)
        else:
            return self.polys1d[dim].derivative_expansion(s, N, K=K)


if __name__ == "__main__":

    pass



import numpy as np


def to_numpy_array(inp):
    """
    Converts floats into numpy arrays.
    """

    if isinstance(inp, float) or isinstance(inp, int):
        return np.asarray([inp])
    else:
        return np.asarray(inp)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()
    
setup(
    name='UncertainSCI',
    version='0.2.1-b0',
    author='UncertainSCI Developers',
    author_email='uncertainsci@sci.utah.edu',
    packages=find_packages(),
    package_dir={'': '.'},
    download_url=r'https://github.com/SCIInstitute/UncertainSCI',
    description=r'A Non-invasive Uncertainty Quantification tool',
    long_description= long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Software Development :: Libraries",
        "Operating System :: Android",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS"
        ],
    license='MIT',
    keywords='Uncertainty Quantification, Simulation',
    url=r'https://sci.utah.edu/sci-software/simulation/uncertainsci.html',
    install_requires=[
        "matplotlib==3.1.3; python_version < '3.8'",
        "matplotlib>=3.1.3; python_version >= '3.8'",
        "numpy==1.15.2; python_version < '3.8'",
        "numpy>=1.21.0; python_version >= '3.8'",
        "scipy==1.4.1; python_version < '3.8'",
        "scipy>=1.4.1; python_version >= '3.8'"
    ]
    )



from packaging import version


def version_lessthan(module, number):
    return version.parse(module.__version__) < version.parse(number)


def version_greaterthan(module, number):
    return version.parse(module.__version__) > version.parse(number)


def version_lessthanorequal(module, number):
    return version.parse(module.__version__) <= version.parse(number)


def version_greaterthanorequal(module, number):
    return version.parse(module.__version__) >= version.parse(number)
